#import "A0bReader.h"
#import "A0bThreadId.h"
#import "../General.bproj/FoundationAdditions.h"
#import "../General.bproj/UniqueString.h"

@implementation A0bReader

static NSString *INT_VALUE = @"Int";
static NSString *HEX_VALUE = @"Hex";
static NSString *TIME_VALUE = @"Date";
static NSString *DOUBLE_VALUE = @"Double";
static NSString *STRING_VALUE = @"String";

- (id)init
{
    [super init];

    eventNames = [[NSMutableDictionary dictionary] retain];
    eventFieldTypes = [[NSMutableDictionary dictionary] retain];
    eventFieldNames = [[NSMutableDictionary dictionary] retain];
    scanner = nil;
    percent = [[NSCharacterSet characterSetWithCharactersInString:@"%"] retain];
    alphanum = [[NSMutableCharacterSet characterSetWithCharactersInString:
        @"%0123456789_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"] retain];
//alphanumericCharacterSet];
//    [alphanum addCharactersInString:@"_"];
    delimiter = [[NSCharacterSet whitespaceAndNewlineCharacterSet] retain];

    return self;
}

- (void)encodeCheckPointWithCoder:(NSCoder *)coder
{
    unsigned int location = bytesScanned + [scanner position];
    [coder encodeValuesOfObjCTypes:"@iI@", &inputFilename, &inputLineNumber, &location, &currentTime];
}

- (void)decodeCheckPointWithCoder:(NSCoder *)coder
{
    id filename;
    unsigned int location;
    id time;

    Assign(scanner, nil);
    Assign(remainsOfLastRead, nil);
    
    [coder decodeValuesOfObjCTypes:"@iI@", &filename, &inputLineNumber, &location, &time];
    bytesScanned = location;
    if (inputFilename) {
        if (![filename isEqual:inputFilename])
            [self raise:@"trying to read check point file of a different trace file"];
    } else {
        [self setInputFilename:filename];
        [filename release];
    }
    [inputFile seekToFileOffset:bytesScanned];
    [self setCurrentTime:time];
    [self readNextLine];
    NSLog(@"%@ decoded check point. offset=%ld, time=%@", NSStringFromClass([self class]), (long)bytesScanned, [currentTime description]);
}

- (void)dealloc
{
    [eventNames release];
    [eventFieldTypes release];
    [eventFieldNames release];
    [inputFilename release];
    [percent release];
    [alphanum release];
    [delimiter release];
    [scanner release];
    [super dealloc];
}

- (void)reset
{
    [eventNames removeAllObjects];
    [eventFieldTypes removeAllObjects];
    [eventFieldNames removeAllObjects];
    Assign(scanner, nil);
}

- (BOOL)readNextLine
{
    if (scanner && ![scanner isAtEnd]) {
        inputLineNumber++;
        return YES;
    }
    return NO;
}

- (void)raise:(NSString *)reason
{
    NSLog(@"A0bReader: '%@' in file '%@', line %d(c%d)", reason, inputFilename, inputLineNumber, [scanner position]);
    [self reset];
    [[NSException exceptionWithName:@"ReadFileException"
                             reason:reason
                           userInfo:
        [NSDictionary dictionaryWithObjectsAndKeys:
            inputFilename, @"File Name",
            [NSNumber numberWithInt:inputLineNumber], @"Line Number",
            nil]
        ] raise];
}


- (void)readTraceFile:(NSString *)fileName
{
    [self reset];
    [self setInputFilename:fileName];
    while([self readNextEvent])
        ;
}

- (NSString *)inputFilename {
    return inputFilename;
}

- (NSDictionary *)setInputFilename:(NSString *)filename
{
    A0Event *event;
    NSData *data;

    Assign(inputFilename, filename);

    data = [NSData dataWithContentsOfMappedFile:filename];
    if (!data)
        [self raise:@"Couldn't open file"];

    Assign(scanner, [DataScanner scannerWithData:data]);

    while([self readNextEventDefinition])
          ;
    
    // raise if no definition found
    if ([eventNames count] == 0)
        [self raise:@"Not an Athapascan0b trace file"];

    // read the first event (with global trace info)
    event = [self readEvent];
    if (event) {
//        [self outputEntity:event];        
        [PTime setInitialTime:[event objectForKey:@"StartTime"]];
        return [NSDictionary dictionaryWithObjectsAndKeys:
            [filename stringByAbbreviatingWithTildeInPath], @"Filename",
            [event objectForKey:@"StartTime"], @"StartTime",
            [event objectForKey:@"EndTime"], @"EndTime",
            [event objectForKey:@"NodeNumber"], @"NodeCount",
            [event objectForKey:@"TraceVersion"], @"TraceVersion",
            nil];
    }
    return nil;
}

- (BOOL)readPercent
{
    int c = [scanner readChar];
    if (c == '%')
        return YES;
    [scanner setPosition:[scanner position] - 1];
    return NO;
}

- (BOOL)readNextEventDefinition
{
    NSNumber *eventId;
    NSString *eventName, *fieldName, *fieldType, *str;
    NSMutableArray *fieldTypeArray = [NSMutableArray array];
    NSMutableArray *fieldNameArray = [NSMutableArray array];

    if (![self readNextLine])
        [self raise:@"Unexpected end of file"];

    if (![self readPercent])
        return NO;

    str = [scanner readString];
    eventName = [scanner readString];
    eventId = [scanner readIntNumber];
    if (![str isEqualToString:@"EventDef"] || !eventName || !eventId)
        [self raise:@"'EventDef name number' expected"];


    [eventNames setObject:[UniqueString stringWithString:eventName]
                   forKey:eventId];
    [fieldNameArray addObject: @"Name"];

    while (YES) {
        if (![self readNextLine])
            [self raise:@"Unexpected end of file"];

        if (![self readPercent])
            [self raise:@"Line should start with a '%'"];

        if (fieldName = [scanner readString]) {
            if ([fieldName isEqualToString:@"EndEventDef"])
                break;

            fieldType = [scanner readString];
            if (!fieldType)
                [self raise:@"Incomplete Line"];

            if ([fieldType isEqualToString:@"int"])
                [fieldTypeArray addObject:INT_VALUE];
            else if ([fieldType isEqualToString:@"hex"])
                [fieldTypeArray addObject:HEX_VALUE];
            else if ([fieldType isEqualToString:@"double"])
                [fieldTypeArray addObject:DOUBLE_VALUE];
            else if ([fieldType isEqualToString:@"date"])
                [fieldTypeArray addObject:TIME_VALUE];
            else if ([fieldType isEqualToString:@"string"])
                [fieldTypeArray addObject:STRING_VALUE];
            else
                [self raise:[NSString stringWithFormat:@"Unrecognised field type '%@'", fieldType]];
            
            [fieldNameArray addObject:fieldName];
        }
    }

    [eventFieldTypes setObject:fieldTypeArray forKey:eventId];
    [eventFieldNames setObject:fieldNameArray forKey:eventId];

    return YES;
}


- (void)readUntilTime:(NSDate *)endTime
{
    A0Event *event;

    do {
        event = [self readEvent];
        if (event)
            [self outputEntity:event];
    } while (event && [currentTime earlierThanDate:endTime]);
}

- (BOOL)readNextEvent
{
    A0Event *event = [self readEvent];
    if (event) {
        [self outputEntity:event];
        return YES;
    }
    return NO;
}

- (A0Event *)readEvent
{
    NSNumber *eventId;
    NSString *eventName;
    A0Event *event;
    static NSMutableArray *valArray;
    NSEnumerator *fieldTypeEnum;
    NSString *type;
    NSNumber *aNumber;
    double aDouble;
    NSString *aString;

    if (!valArray)
        valArray = [NSMutableArray new];

redo:
    eventId = [scanner readIntNumber];
    if (!eventId)
        return nil;
    
    eventName = [eventNames objectForKey:eventId];
    if (eventName == nil) {
        NSLog(@"A0bReader: ignoring unrecognized event number %@ at line %d\n",
              eventId, inputLineNumber);
        [self readNextLine];
        goto redo;
    }
    [valArray addObject:eventName];

    fieldTypeEnum = [[eventFieldTypes objectForKey:eventId] objectEnumerator];
    while(type = [fieldTypeEnum nextObject]) {
        if (type == INT_VALUE) {
            aNumber = [scanner readIntNumber];
            [valArray addObject:aNumber];
        } else if (type == DOUBLE_VALUE) {
            aNumber = [scanner readDoubleNumber];
            [valArray addObject:aNumber];
        } else if (type == TIME_VALUE) {
            aDouble = [scanner readDouble];
            [valArray addObject:[PTime dateWithTimeIntervalSinceReferenceDate:aDouble]];
        } else if (type == HEX_VALUE || type == STRING_VALUE) {
            // string can be delimited by ""
            aString = [scanner readString];
            [valArray addObject:aString];
        } else
            [self raise: @"Unknown data type"];
    }
    [self readNextLine];

    event = [A0Event eventWithObjects:valArray
                              forKeys:[eventFieldNames objectForKey:eventId]];
    [valArray removeAllObjects];

    [self setCurrentTime:[event time]];

    return event;
}

- (NSDate *)currentTime
{
    return currentTime;
}

- (void)setCurrentTime:(NSDate *)time
{
    Assign(currentTime, time);
}
@end
