/* A0bMutex.m created by benhur on 27.apr.1998 */

#import "A0bMutex.h"
#import "A0State.h"
#import "A0Link.h"
#import "../General.bproj/UniqueString.h"

PajeContainerType *A0MutexEntityType;

@implementation A0bMutex

+ (void)initialize
{
    if (self == [A0bMutex class]) {
        A0MutexEntityType = [[PajeContainerType alloc] initWithName:[UniqueString stringWithString:@"Mutex"] containerType:[A0bNode entityType]];
    }
    [super initialize];
}

+ (PajeContainerType *)entityType { return A0MutexEntityType; }

- (id)initWithId:(id)ident node:(A0bNode *)node
{
    self = [super initWithType:A0MutexEntityType
                          name:ident
                     container:node];
    if (self != nil) {
        simulator = nil;
        lastEvent = nil;
        currentState = nil;
        waitingStates = [[NSMutableDictionary alloc] init];
        holdingState = nil;
        lastUnlockEvent = nil;
        creationTime = nil;
        destructionTime = nil;
    }
    return self;
}

- (void)dealloc
{
    // doesn' release node, it has not been retained in init
    [lastEvent release];
    [currentState release];
    [waitingStates release];
    [holdingState release];
    [lastUnlockEvent release];
    [creationTime release];
    [destructionTime release];
    [super dealloc];
}

+ (A0bMutex *)mutexWithId:(id)ident node:(A0bNode *)n
{
    return [[[self alloc] initWithId:ident node:n] autorelease];
}

- (A0bNode *)node { return (A0bNode *)container; }

- (void)setSimulator:(A0bSimul *)simul
{
    NSAssert1(!simulator,
              @"Simulator being set more than once in semaphore %@",
              [self description]);
    simulator = simul;
}

- (void)setLastEvent:(A0Event *)event
{
    [event retain];
    [lastEvent release];
    lastEvent = event;
}

- (void)setLockEvent:(A0Event *)event
{
    [event retain];
    [lockEvent release];
    lockEvent = event;
}

- (void)setLastUnlockEvent:(A0Event *)event
{
    [event retain];
    [lastUnlockEvent release];
    lastUnlockEvent = event;
}

- (void)unlockWithEvent:(A0Event *)event
{
    NSAssert1(holdingState != nil,
              @"unlocking an unlocked mutex with event: %@", event);
    // complete and output thread mutex state
    [holdingState setEndEvent:event];
//    [simulator outputEntity:holdingState];
    [holdingState release];
    holdingState = nil;

    [self setLockEvent:nil];
}

- (void)a0NewMutex:(A0Event *)event
{
    NSAssert1(simulator,
              @"Simulator not set in mutex for event %@",
              [event description]);
    NSAssert1(!lastEvent,
              @"Mutex being recreated with event %@",
              [event description]);

    creationTime = [[event time] retain];
    
    currentState = [[A0MutexState unlockedMutexStateFromEvent:event mutex:self] retain];

    [self setLastEvent:event];

    [simulator outputEntity:currentState];
    [simulator outputEntity:
        [A0Link newMutexLinkFromEvent:event toState:currentState]];
}

- (void)a0LockMutex:(A0Event *)event
{
    A0bThread *thread = [simulator threadForEvent:event];
    A0MutexThreadState *waitingState;
    A0State *threadState;

    NSAssert1(thread, @"Unknown thread for event %@", event);

    // output the current state
    [currentState setEndEvent:event];
//    [simulator outputEntity:currentState];
    [currentState release];
    // and generate a new one
    currentState = [[A0MutexState lockedMutexStateFromEvent:event
                                                      mutex:self] retain];

    [thread blockWithEvent:event
           unblockSelector:@selector(thread:unblockedWithEvent:)
                    target:self];

    threadState = [thread currentState];
    [threadState addRelatedEntity:currentState];
    [currentState addRelatedEntity:threadState];

    // generate a waiting thread mutex state for the thread.
    waitingState = [A0MutexThreadState waitingMutexThreadStateFromEvent:event
                                                                  mutex:self];
    [threadState addRelatedEntity:waitingState];
    [currentState addRelatedEntity:waitingState];
    [waitingState addRelatedEntity:threadState];
    [waitingState addRelatedEntity:currentState];
    [waitingStates setObject:waitingState forKey:thread];

    [self setLastEvent:event];
    [simulator outputEntity:currentState];
    [simulator outputEntity:waitingState];
}

- (void)a0UnlockMutex:(A0Event *)event
{
    [currentState setEndEvent:event];
//    [simulator outputEntity:currentState];
    [currentState release];
    
    if ([waitingStates count] > 0) {
        // some thread is waiting for this mutex. Set the event 
        // to link to when it unlocks (in -thread:unblockedWithEvent:)
        [self setLastUnlockEvent:event];
        currentState = [[A0MutexState lockedMutexStateFromEvent:event mutex:self] retain];
    } else {
        currentState = [[A0MutexState unlockedMutexStateFromEvent:event mutex:self] retain];
    }

    [self unlockWithEvent:event];
    [self setLastEvent:event];
    [simulator outputEntity:currentState];
}

- (void)a0DisposeMutex:(A0Event *)event
{
    destructionTime = [[event time] retain];
    [currentState setEndEvent:event];
//    [simulator outputEntity:currentState];
    [currentState release];
    currentState = nil;

    if ([waitingStates count]) {
        [simulator error:@"disposing mutex with blocked threads %@", event];
    }
    [self setLastEvent:event];
}

- (void)a0WaitCondition:(A0Event *)event
{
    A0bThread *thread = [simulator threadForEvent:event];
    A0State *threadState;

    // unlock mutex
    [currentState setEndEvent:event];
//    [simulator outputEntity:currentState];
    [currentState release];

    if ([waitingStates count] > 0) {
        // some thread is waiting for this mutex. Set the event
        // to link to when it unlocks (in -thread:unblockedWithEvent:)
        [self setLastUnlockEvent:event];
        currentState = [[A0MutexState lockedMutexStateFromEvent:event mutex:self] retain];
    } else {
        currentState = [[A0MutexState unlockedMutexStateFromEvent:event mutex:self] retain];
    }

    [self unlockWithEvent:event];


    // lock mutex
    NSAssert1(thread, @"Unknown thread for event %@", event);

    [thread blockWithEvent:event
           unblockSelector:@selector(thread:unblockedWithEvent:)
                    target:self];

    threadState = [thread currentState];
    [threadState addRelatedEntity:currentState];
    [currentState addRelatedEntity:threadState];

//    [blockedThreads addObject:thread];

    [self setLastEvent:event];

    [simulator outputEntity:currentState];
}

- (void)thread:(A0bThread *)thread unblockedWithEvent:(A0Event *)event
// a thread that was blocked waiting for this mutex has unblocked.
// it now holds the mutex.
{
    A0MutexThreadState *waitingState;

    NSAssert2(holdingState == nil,
              @"mutex held by thread %@ being locked with event: %@",
              [holdingState thread], event);

    // complete the waiting mutex thread state and output it
    waitingState = [waitingStates objectForKey:thread];
    if (waitingState) {
        [waitingState setEndEvent:event];
        if (lastUnlockEvent)
            [waitingState addRelatedEntity:lastUnlockEvent];
//        [simulator outputEntity:waitingState];
        [waitingStates removeObjectForKey:thread];
    }

//    [currentState removeRelatedEntity:[thread currentState]];

    // make a link between the last unlocking event and this event.
    if (lastUnlockEvent) {
        [simulator outputEntity:
            [A0Link unlockMutexLinkFromEvent:lastUnlockEvent
                                     toEvent:event]];
        [self setLastUnlockEvent:nil];
    }

    // thread now holds the lock. generate its holding thread mutex state
    // it will be completed (end time) in unlockWithEvent:
    holdingState = [[A0MutexThreadState holdingMutexThreadStateFromEvent:event mutex:self] retain];
    [simulator outputEntity:holdingState];
}

- (NSDate *)startTime {
    return creationTime;
}
- (NSDate *)endTime {
    if (destructionTime != nil)
        return destructionTime;
    return [NSDate distantFuture];
}

// NSCoding Protocol
- (void)encodeWithCoder:(NSCoder *)coder
{
    [super encodeWithCoder:coder];

//    [coder encodeConditionalObject:simulator];
    [coder encodeObject:lastEvent];
    [coder encodeObject:currentState];
    [coder encodeObject:waitingStates];
    [coder encodeObject:holdingState];
    [coder encodeObject:lastUnlockEvent];
    [coder encodeObject:creationTime];
    [coder encodeObject:destructionTime];
}

- (id)initWithCoder:(NSCoder *)coder
{
    self = [super initWithCoder:coder];

//    simulator = [coder decodeObject];
    lastEvent = [[coder decodeObject] retain];
    currentState = [[coder decodeObject] retain];
    waitingStates = [[coder decodeObject] retain];
    holdingState = [[coder decodeObject] retain];
    lastUnlockEvent = [[coder decodeObject] retain];
    creationTime = [[coder decodeObject] retain];
    destructionTime = [[coder decodeObject] retain];

    return self;
}
@end
