/* A0bSimul-Comm.m created by benhur on Mon 21-Apr-1997 */

#import "A0bSimul.h"

@implementation A0bSimul (Communications)
/*
 * --------
 * Messages
 * --------
 * This is not straightforward...
 * We must match four different events to make a communication:
 * the a0Send, a0Receive (to get node and thread identification)
 * and a0SendCompletion and a0ReceiveCompletion (to get timing
 * information and source node, in the case of a0AnySource).
 * This match is done in two phases.
 * First, the Send is matched with the SendCompletion
 * and the Receive with the ReceiveCompletion;
 * then, the completed Send and Receive are matched.
 *
 * The first match is done by matching the SerialNo field 
 * of both the completion and the communication events,
 * that's guaranteed to be unique by Athapascan.
 * In the trace file (ordered by time), the completion 
 * event always comes after the corresponding communication event.
 * The incomplete communication events are stored in
 * incompleteSends and incompleteReceives of each node, 
 * with SerialNo as key.
 *
 * The match of a Send and a Receive is done based in the FIFO
 * properties of Athapascan, that garantees that the messages
 * from one node to another are sent in the same order as they
 * are posted. But their timestamps are not garanteed keep in that
 * same order; for this, there is a SerialNo field in each 
 * communication event, whose order must be observed when matching
 * these events.
 * For the match, we have a list of completed Sends (and another for
 * receives), ordered by SerialNo, that is searched serially.
 * The problem is that not all completed events can be put on that
 * list, because, due to ordering problems, we could have one event
 * to complete before another that has a smaller SerialNo
 * that are concurrent (have the same source, destin (or any), port
 * and tag) and in that case we could have an erroneous match.
 * To solve this, we have an intermediate list (in fact, two for each
 * node, one for sends and another for receives) where all events are
 * put before going to the completed list.
 * Each time one event in this list completes, it's compared against
 * all events before it in the list, to see if there's one in which
 * it depends. If there is, nothing is done.
 * If there is not, it's put in the completed list and compared with
 * the events that were after it, searching for events that were depending
 * on it and releasing them, until an event in which it would also depend.
 */

// Protocols for Pipes
#define A0AsyncPipe		0x000
#define A0SyncPipe		0x001
#define A0FlowPipe		0x008
#define A0OutputPipe		0x010
#define A0InputPipe		0x020


- (void)a0Send:(A0bEvent *)event
{
    A0bNode *node = [nodes objectForKey:[event node]];

    NSAssert1(node, @"Unknown node in event: %@", event);

    // may be a blocking event (if it has EndTime).
    [self blockThreadWithEvent: event blockingState:PThreadBlockedSend];

    // put event in pendingSends, that is ordered by SerialNo
    [[node pendingSends] addObject:event];
    // see if there is a match
    [self verifyMatchesSendSendCompletionInNode:node];
}

- (void)a0SendCompletion:(A0bEvent *)event
{
    A0bNode *node = [nodes objectForKey:[event node]];

    NSAssert1(node, @"Unknown node in event: %@", event);

    // put event in pendingSendCompletions, indexed by SerialNo
    [[node pendingSendCompletions] setObject:event forKey:[event serialNo]];
    // see if there is a match
    [self verifyMatchesSendSendCompletionInNode:node];
}

- (void)verifyMatchesSendSendCompletionInNode:(A0bNode *)sourceNode
{
    PSortedArray *pendingSends = [sourceNode pendingSends];
    PSortedArray *pendingSendCompletions = [sourceNode pendingSendCompletions];
    A0bEvent *sendEvent, *completionEvent;

    while (([pendingSends count] > 0) && ([pendingSendCompletions count] > 0)) {
        sendEvent = [pendingSends objectAtIndex:0];
        completionEvent = [pendingSendCompletions objectAtIndex:0];
        if ([[sendEvent serialNo] isEqual:[completionEvent serialNo]]) {
            [self matchSendEvent:sendEvent withCompletion:completionEvent];
            [pendingSends removeObjectAtIndex:0];
            [pendingSendCompletions removeObjectAtIndex:0];
        } else
            break;
    }
}

- (void)matchSendEvent:(A0bEvent *)sendEvent
        withCompletion:(A0bEvent *)completionEvent
{
    NSEnumerator *enumer;
    A0bComm *comm;
    A0bNode *sourceNode = [nodes objectForKey:[sendEvent node]];

    NSAssert1(sourceNode, @"Unknown node in event: %@", sendEvent);

    // test to see if the corresponding receive has already been processed.
    enumer = [[sourceNode pendingReceiveComms] objectEnumerator];
    while (comm = [enumer nextObject]) {
        A0bEvent *recvEvent = [comm receiveEvent];

        NSAssert1(recvEvent,
                  @"receive event not found in communication:%@", comm);
        
        if ([[sendEvent destNode] isEqual:[recvEvent node]]
            && [[sendEvent port] isEqual:[recvEvent port]]
            && [[sendEvent tag] isEqual:[recvEvent tag]]) {

            [comm setSendEvent:sendEvent];
            [comm setSendCompletionEvent:completionEvent];
            [comm setSourceService:[self serviceForEvent:sendEvent]];

            [[self outputNum:COMM_OUTPUT_PORT] setPortValue:comm];
            [[sourceNode pendingReceiveComms] removeObjectIdenticalTo:comm];
            break;
        }
    }

    // if receive wasn't found, insert an incomplete comm in receiver's
    // pending send communications queue
    if (!comm) {
        A0bNode *destNode = [nodes objectForKey:[sendEvent destNode]];

        NSAssert1(destNode, @"Unknown destination node in event: %@", sendEvent);

        comm = [A0bComm commWithName:PCommMessage];

        [comm setSendEvent:sendEvent];
        [comm setSendCompletionEvent:completionEvent];
        [comm setSourceService:[self serviceForEvent:sendEvent]];

        [[destNode pendingSendComms] addObject:comm];
    }
}




- (void)a0Receive:(A0bEvent *)event
{
    A0bNode *node = [nodes objectForKey:[event node]];

    if (!node) {
        [self error:@"Event in unknown node: %@", event];
        return;
    }

    // may be a blocking event (if it has EndTime).
    [self blockThreadWithEvent:event blockingState:PThreadBlockedReceive];

    // put event in pendingReceives, that's ordered by SerialNo
    [[node pendingReceives] addObject:event];
}

- (void)a0ReceiveCompletion:(A0bEvent *)event
{
    id sourceNodeId, otherSourceNodeId;
    id serialNo;
    A0bNode *destNode;
    A0bEvent *receiveEvent, *otherReceive, *completionEvent;
    id destNodeId;
    id port;
    id tag;
    id size;
    PSortedArray *pendingReceives;
    NSMutableDictionary *pendingReceiveCompletions;
    BOOL conflictExists;
    int receiveIndex, index;

    destNodeId = [event node];
    destNode = [nodes objectForKey:destNodeId];
    if (!destNode) {
        [self error:@"Event in unknown node: %@", event];
        return;
    }
    serialNo = [event serialNo];

    // search the corresponding receive event
    pendingReceives = [destNode pendingReceives];
    receiveIndex = [pendingReceives indexOfObjectWithValue:serialNo];
    if (receiveIndex == NSNotFound) {
//        [self error:@"Completion of unknown communication: %@", event];
        [event setObject:@"Completion of unknown communication" forKey:@"Error"];
        [event setColor:[NSColor redColor]];
        return;
    }

    // if there's a pending receive with the same origin, port and tag and
    // a smaller serial number than this one, we can't go any further.
    receiveEvent = [pendingReceives objectAtIndex:receiveIndex];
    sourceNodeId = [event objectForKey:@"SourceNode"];
    port = [receiveEvent objectForKey:@"Port"];
    tag = [receiveEvent objectForKey:@"Tag"];

    conflictExists = NO;
    for (index = 0; index < receiveIndex; index++) {
        otherReceive = [pendingReceives objectAtIndex:index];
        otherSourceNodeId = [otherReceive objectForKey:@"SourceNode"];
        if ((!otherSourceNodeId
             ||  [sourceNodeId isEqual:otherSourceNodeId])
            && [port isEqual:[otherReceive objectForKey:@"Port"]]
            && [tag isEqual:[otherReceive objectForKey:@"Tag"]]) {
            conflictExists = YES;
            break;
        }
    }
    pendingReceiveCompletions = [destNode pendingReceiveCompletions];
    if (conflictExists) {
        [pendingReceiveCompletions setObject:event forKey:serialNo];
        return;
    }

    // this receive is free to search it's send!
    // let's also free everyone that's waiting for it.

    completionEvent = event;
    index = receiveIndex;
    do {
        NSEnumerator *enumer;
        A0bComm *comm;
        
        // test to see if the corresponding send has already been processed.
        enumer = [[destNode pendingSendComms] objectEnumerator];
        while (comm = [enumer nextObject]) {
            A0bEvent *sendEvent = [[comm objectForKey:@"RelatedEntities"]
                                         objectForKey:@"SendEvent"];

            if ([sourceNodeId isEqual: [sendEvent node]]
                && [port isEqual: [sendEvent objectForKey:@"Port"]]
                && [tag isEqual: [sendEvent objectForKey:@"Tag"]]) {

                [comm setObject:[completionEvent time] forKey:@"EndTime"];
                [comm setObject:[receiveEvent thread] forKey:@"DestThread"];
                [comm setObject:[self serviceForEvent:receiveEvent] forKey:@"DestService"];
                size = [completionEvent objectForKey:@"Size"];
                if (size) [comm setObject:size forKey:@"Size"];

                [[comm objectForKey:@"RelatedEntities"] addEntriesFromDictionary:
                    [NSDictionary dictionaryWithObjectsAndKeys:
                        receiveEvent, @"ReceiveEvent",
                        completionEvent, @"ReceiveCompletionEvent",
                        nil]];

                [[self outputNum:COMM_OUTPUT_PORT] setPortValue:comm];
                [[destNode pendingSendComms] removeObjectIdenticalTo:comm];
                break;
            }
        }
        
        // if send wasn't found, insert an incomplete comm in sender's
        // pending receive communications queue
        if (!comm) {
            A0bNode *sourceNode = [nodes objectForKey:sourceNodeId];
            if (!sourceNode) {
                [self error:@"Receive from unknown node: %@", event];
                return;
            }

            comm = [A0bComm commWithName:PCommMessage];
            [comm setObject:[completionEvent time] forKey:@"EndTime"];
            [comm setObject:sourceNodeId forKey:@"SourceNode"];
            [comm setObject:destNodeId forKey:@"DestNode"];
            [comm setObject:[receiveEvent thread] forKey:@"DestThread"];
            [comm setObject:[self serviceForEvent:receiveEvent] forKey:@"DestService"];
            size = [completionEvent objectForKey:@"Size"];
            if (size) [comm setObject:size forKey:@"Size"];
            [comm setObject:[NSMutableDictionary dictionaryWithObjectsAndKeys:
                                   receiveEvent, @"ReceiveEvent",
                                   completionEvent, @"ReceiveCompletionEvent",
                                   nil]
                     forKey:@"RelatedEntities"];

            [[sourceNode pendingReceiveComms] addObject:comm];
        }

        // it's not pending anymore!
        [pendingReceives removeObjectAtIndex:index];
        if (completionEvent != event)
            [pendingReceiveCompletions removeObjectForKey:serialNo];

        // search a receive that was pending on us.
        // if it is completed, we'll process it and continue (in this case,
        // completionEvent will not be nil).
        // if it is not completed, we stop, as any later receive is also
        // depending on this incomplete one and can't be processed now.
        completionEvent = nil;
        while(index < [pendingReceives count]) {
            receiveEvent = [pendingReceives objectAtIndex:index];
            otherSourceNodeId = [receiveEvent objectForKey:@"SourceNode"];
            if ((!otherSourceNodeId
                 ||  [sourceNodeId isEqual:otherSourceNodeId])
                && [port isEqual:[receiveEvent objectForKey:@"Port"]]
                && [tag isEqual:[receiveEvent objectForKey:@"Tag"]]) {
                serialNo = [receiveEvent serialNo];
                completionEvent = [pendingReceiveCompletions
                    objectForKey:serialNo];
                break;
            } else
                index++;
        }
        
    } while (completionEvent);
}



- (void)a0WaitRequest:(A0bEvent *)event
{
    [self blockThreadWithEvent:event blockingState:PThreadBlockedWaitRequest];
}


- (void)a0NewPipe:(A0bEvent *)event
{
    A0bComm *comm;
    id nodeId = [event node];
    id newThreadId = [event objectForKey:@"OtherThread"];
    id time = [event time];
    int protocol = [[event objectForKey:@"Protocol"] intValue];

    [event setObject:[NSString stringWithFormat:@"%d (%@%@%@%@)", protocol,
             (protocol & A0SyncPipe) ? @"Sync" : @"Async",
             (protocol & A0FlowPipe) ? @", Flow" : @"",
             (protocol & A0OutputPipe) ? @", Output" : @"",
             (protocol & A0InputPipe) ? @", Input" : @""] forKey:@"Protocol"];

    if (!(protocol & A0SyncPipe)) {
        // Assynchronous pipes create a thread to treat them; no join--detach
        [[self startThread:newThreadId
                    atNode:nodeId
                    atTime:[event time] withEvent:event] detach];

        comm = [A0bComm commWithName:PCommNewPipe];
        [comm setObject:nodeId forKey:@"SourceNode"];
        [comm setObject:nodeId forKey:@"DestNode"];
        [comm setObject:[event thread] forKey:@"SourceThread"];
        [comm setObject:[self serviceForEvent:event] forKey:@"SourceService"];
        [comm setObject:newThreadId forKey:@"DestThread"];
        [comm setObject:[self serviceForNode:nodeId thread:newThreadId] forKey:@"DestService"];
        [comm setObject:time forKey:@"StartTime"];
        [comm setObject:time forKey:@"EndTime"];
        [comm setObject:[NSMutableArray arrayWithObject:event]
                 forKey:@"RelatedEntities"];
        [[self outputNum:COMM_OUTPUT_PORT] setPortValue:comm];
    }

}
@end
