/* A0bSimul-Comm.m created by benhur on Mon 21-Apr-1997 */

#import "A0bSimul.h"

@implementation A0bSimul (Communications)
/*
 * --------
 * Messages
 * --------
 * This is not straightforward...
 * We must match four different events to make a communication:
 * the a0Send, a0Receive (to get node and thread identification)
 * and a0SendCompletion and a0ReceiveCompletion (to get timing
 * information and source node, in the case of a0AnySource).
 * This match is done in two phases.
 * First, the Send is matched with the SendCompletion
 * and the Receive with the ReceiveCompletion;
 * then, the completed Send and Receive are matched.
 *
 * The first match is done by matching the SerialNo field 
 * of both the completion and the communication events,
 * that's guaranteed to be unique by Athapascan.
 * In the trace file (ordered by time), the completion 
 * event always comes after the corresponding communication event.
 * The incomplete communication events are stored in
 * incompleteSends and incompleteReceives of each node, 
 * with SerialNo as key.
 *
 * The match of a Send and a Receive is done based in the FIFO
 * properties of Athapascan, that garantees that the messages
 * from one node to another are sent in the same order as they
 * are posted. But their timestamps are not garanteed keep in that
 * same order; for this, there is a SerialNo field in each 
 * communication event, whose order must be observed when matching
 * these events.
 * For the match, we have a list of completed Sends (and another for
 * receives), ordered by SerialNo, that is searched serially.
 * The problem is that not all completed events can be put on that
 * list, because, due to ordering problems, we could have one event
 * to complete before another that has a smaller SerialNo
 * that are concurrent (have the same source, destin (or any), port
 * and tag) and in that case we could have an erroneous match.
 * To solve this, we have an intermediate list (in fact, two for each
 * node, one for sends and another for receives) where all events are
 * put before going to the completed list.
 * Each time one event in this list completes, it's compared against
 * all events before it in the list, to see if there's one in which
 * it depends. If there is, nothing is done.
 * If there is not, it's put in the completed list and compared with
 * the events that were after it, searching for events that were depending
 * on it and releasing them, until an event in which it would also depend.
 */

// Protocols for Pipes
#define A0AsyncPipe		0x000
#define A0SyncPipe		0x001
#define A0FlowPipe		0x008
#define A0OutputPipe		0x010
#define A0InputPipe		0x020


- (void)a0Send:(A0Event *)event
{
    A0bNode *node = [self nodeForEvent:event];

    if (!node) {
        [self error:@"Event in unknown node: %@", event];
        return;
    }

    // may be a blocking event (if it has EndTime).
    if ([[event name] characterAtIndex:2] != 'I')
        [self blockThreadWithEvent: event];

    if (traceVersion == 1)
    {
        id req = [event objectForKey:@"Request"];
        [[node reallyPendingSends] setObject:event forKey:req];
        return;
    }
    
    // put event in pendingSends, that is ordered by SerialNo
    [[node pendingSends] addObject:event];
}


- (void)a0_SendCompletion:(A0Event *)event
{
    id sourceNodeId;
    id serialNo;
    A0bNode *sourceNode;
    A0Event *sendEvent, *otherSend, *completionEvent;
    id destNodeId;
    id port;
    id tag;
    PSortedArray *pendingSends;
    NSMutableDictionary *pendingSendCompletions;
    BOOL conflictExists;
    int sendIndex, index;

    sourceNodeId = [event node];
    sourceNode = [nodes objectForKey:sourceNodeId];
    if (!sourceNode) {
        [self error:@"Event in unknown node: %@", event];
        return;
    }
    serialNo = [event serialNo];

    if (traceVersion == 1) {
        id req = [event objectForKey:@"Request"];
        sendEvent = [[sourceNode reallyPendingSends] objectForKey:req];
        if (sendEvent) {
            [sendEvent setObject:serialNo forKey:@"SerialNo"];
            [[sourceNode pendingSends] addObject:sendEvent];
            [[sourceNode reallyPendingSends] removeObjectForKey:req];
        }
    }
        
    // search the corresponding send event
    pendingSends = [sourceNode pendingSends];
    sendIndex = [pendingSends indexOfObjectWithValue:serialNo];
    if (sendIndex == NSNotFound) {
        [event setObject:@"Completion of unknown communication" forKey:@"Error"];
        [event setColor:[NSColor redColor]];
        return;
    }

    // if there's a pending send with the same destin, port and tag and
    // a smaller serial number than this one, we can't go any further.
    sendEvent = [pendingSends objectAtIndex:sendIndex];
    destNodeId = [sendEvent objectForKey:@"DestNode"];
    port = [[sendEvent objectForKey:@"Port"] retain];
    tag = [[sendEvent objectForKey:@"Tag"] retain];

    [event setThread:[sendEvent thread]];

    conflictExists = NO;
    for (index = 0; index < sendIndex; index++) {
        otherSend = [pendingSends objectAtIndex:index];
        if ([destNodeId isEqual:[otherSend objectForKey:@"DestNode"]]
            && [port isEqual:[otherSend objectForKey:@"Port"]]
            && [tag isEqual:[otherSend objectForKey:@"Tag"]]) {
            conflictExists = YES;
            break;
        }
    }
    pendingSendCompletions = [sourceNode pendingSendCompletions];
    if (conflictExists) {
        [pendingSendCompletions setObject:event forKey:serialNo];
        return;
    }

    // this send is free to search it's receive!
    // let's also free everyone that's waiting for it.

//    NSLog(@"searching receive for:%@ %@ %@", destNodeId, port, tag);
    
    completionEvent = event;
    index = sendIndex;
    do {
        NSEnumerator *enumer;
        A0Comm *comm;
        
        // test to see if the corresponding receive has already been processed.
        enumer = [[sourceNode pendingReceiveComms] objectEnumerator];
        while (comm = [enumer nextObject]) {

//            NSLog(@"trying receive for:%@ %@ %@", [comm destNodeId], [comm port], [comm tag]);
            if ([destNodeId isEqual:[comm destNodeId]]
                && [port isEqual:[comm port]]
                && [tag isEqual:[comm tag]]) {

                [comm setSendEvent:sendEvent];
                [comm setSendCompletionEvent:completionEvent];

                [self outputEntity:comm];
                [[sourceNode pendingReceiveComms] removeObjectIdenticalTo:comm];
//                NSLog(@"FOUND");
                break;
            }
        }
        
        // if receive wasn't found, insert an incomplete comm in receiver's
        // pending send communications queue
        if (!comm) {
            A0bNode *destNode = [nodes objectForKey:destNodeId];
            if (!destNode) {
                [self error:@"a0Send to unknown node: %@", event];
                return;
            }

//            NSLog(@"NOT FOUND");
//            NSLog(@"storing pending send:%@->%@", sourceNode, destNode);
            comm = [A0Comm communication];
            [comm setSendEvent:sendEvent];
            [comm setSendCompletionEvent:completionEvent];

            [[destNode pendingSendComms] addObject:comm];
        }

        // it's not pending anymore!
        if (completionEvent != event)
            [pendingSendCompletions removeObjectForKey:serialNo];
        [pendingSends removeObjectAtIndex:index];

        // search a send that was pending on us.
        // if it is completed, we'll process it and continue (in this case,
        // completionEvent will not be nil).
        // if it is not completed, we stop, as any later send is also
        // depending on this incomplete one and can't be processed now.
        completionEvent = nil;
        while(index < [pendingSends count]) {
            sendEvent = [pendingSends objectAtIndex:index];
            if ([destNodeId isEqual:[sendEvent objectForKey:@"DestNode"]]
                    && [port isEqual:[sendEvent objectForKey:@"Port"]]
                    && [tag isEqual:[sendEvent objectForKey:@"Tag"]]) {
                serialNo = [sendEvent serialNo];
                completionEvent = [pendingSendCompletions
                    objectForKey:serialNo];
                break;
            } else
                index++;
        }
        
    } while (completionEvent);
    [port release];
    [tag release];
}




- (void)a0Receive:(A0Event *)event
{
    A0bNode *node = [self nodeForEvent:event];

    if (!node) {
        [self error:@"Event in unknown node: %@", event];
        return;
    }

    // may be a blocking event (if it has EndTime).
    if ([[event name] characterAtIndex:2] != 'I')
        [self blockThreadWithEvent: event];

    if (traceVersion == 1)
    {
        id req = [event objectForKey:@"Request"];
        [[node reallyPendingReceives] setObject:event forKey:req];
        return;
    }
    
    // put event in pendingReceives, that's ordered by SerialNo
    [[node pendingReceives] addObject:event];
}

- (void)a0_ReceiveCompletion:(A0Event *)event
{
    id sourceNodeId, otherSourceNodeId;
    id serialNo;
    A0bNode *destNode;
    A0Event *receiveEvent, *otherReceive, *completionEvent;
    id destNodeId;
    id port;
    id tag;
    PSortedArray *pendingReceives;
    NSMutableDictionary *pendingReceiveCompletions;
    BOOL conflictExists;
    int receiveIndex, index;

    destNodeId = [event node];
    destNode = [nodes objectForKey:destNodeId];
    if (!destNode) {
        [self error:@"Event in unknown node: %@", event];
        return;
    }
    serialNo = [event serialNo];

    if (traceVersion == 1) {
        id req = [event objectForKey:@"Request"];
        receiveEvent = [[destNode reallyPendingReceives] objectForKey:req];
        if (receiveEvent) {
            [receiveEvent setObject:serialNo forKey:@"SerialNo"];
            [[destNode pendingReceives] addObject:receiveEvent];
            [[destNode reallyPendingReceives] removeObjectForKey:req];
        }
    }

    // search the corresponding receive event
    pendingReceives = [destNode pendingReceives];
    receiveIndex = [pendingReceives indexOfObjectWithValue:serialNo];
    if (receiveIndex == NSNotFound) {
//        [self error:@"Completion of unknown communication: %@", event];
        [event setObject:@"Completion of unknown communication" forKey:@"Error"];
        [event setColor:[NSColor redColor]];
        return;
    }

    // if there's a pending receive with the same origin, port and tag and
    // a smaller serial number than this one, we can't go any further.
    receiveEvent = [pendingReceives objectAtIndex:receiveIndex];
    sourceNodeId = [event objectForKey:@"SourceNode"];
    port = [[receiveEvent objectForKey:@"Port"] retain];
    tag = [[receiveEvent objectForKey:@"Tag"] retain];

    [event setThread:[receiveEvent thread]];
    
    conflictExists = NO;
    for (index = 0; index < receiveIndex; index++) {
        otherReceive = [pendingReceives objectAtIndex:index];
        otherSourceNodeId = [otherReceive objectForKey:@"SourceNode"];
        if ((!otherSourceNodeId
             ||  [sourceNodeId isEqual:otherSourceNodeId])
            && [port isEqual:[otherReceive objectForKey:@"Port"]]
            && [tag isEqual:[otherReceive objectForKey:@"Tag"]]) {
            conflictExists = YES;
            break;
        }
    }
    pendingReceiveCompletions = [destNode pendingReceiveCompletions];
    if (conflictExists) {
        [pendingReceiveCompletions setObject:event forKey:serialNo];
        return;
    }

    // this receive is free to search it's send!
    // let's also free everyone that's waiting for it.

//    NSLog(@"searching send for:%@ %@ %@", sourceNodeId, port, tag);
    
    completionEvent = event;
    index = receiveIndex;
    do {
        NSEnumerator *enumer;
        A0Comm *comm;
        
        // test to see if the corresponding send has already been processed.
        enumer = [[destNode pendingSendComms] objectEnumerator];
        while (comm = [enumer nextObject]) {

//            NSLog(@"trying send for:%@ %@ %@", [comm sourceNodeId], [comm port], [comm tag]);
            if ([sourceNodeId isEqual:[comm sourceNodeId]]
                && [port isEqual:[comm port]]
                && [tag isEqual:[comm tag]]) {

                [comm setRecvEvent:receiveEvent];
                [comm setRecvCompletionEvent:completionEvent];

                [self outputEntity:comm];
                [[destNode pendingSendComms] removeObjectIdenticalTo:comm];
//                NSLog(@"FOUND");
                break;
            }
        }
        
        // if send wasn't found, insert an incomplete comm in sender's
        // pending receive communications queue
        if (!comm) {
            A0bNode *sourceNode = [nodes objectForKey:sourceNodeId];
            if (!sourceNode) {
                [self error:@"Receive from unknown node: %@", event];
                return;
            }

//            NSLog(@"NOT FOUND");
//            NSLog(@"storing pending receive:%@->%@", sourceNode, destNode);
            comm = [A0Comm communication];
            [comm setRecvEvent:receiveEvent];
            [comm setRecvCompletionEvent:completionEvent];
            
            [[sourceNode pendingReceiveComms] addObject:comm];
        }

        // it's not pending anymore!
        if (completionEvent != event)
            [pendingReceiveCompletions removeObjectForKey:serialNo];
        [pendingReceives removeObjectAtIndex:index];

        // search a receive that was pending on us.
        // if it is completed, we'll process it and continue (in this case,
        // completionEvent will not be nil).
        // if it is not completed, we stop, as any later receive is also
        // depending on this incomplete one and can't be processed now.
        completionEvent = nil;
        while(index < [pendingReceives count]) {
            receiveEvent = [pendingReceives objectAtIndex:index];
            otherSourceNodeId = [receiveEvent objectForKey:@"SourceNode"];
            if ((!otherSourceNodeId
                 ||  [sourceNodeId isEqual:otherSourceNodeId])
                && [port isEqual:[receiveEvent objectForKey:@"Port"]]
                && [tag isEqual:[receiveEvent objectForKey:@"Tag"]]) {
                serialNo = [receiveEvent serialNo];
                completionEvent = [pendingReceiveCompletions
                    objectForKey:serialNo];
                break;
            } else
                index++;
        }
        
    } while (completionEvent);
    [port release];
    [tag release];
}



- (void)a0WaitRequest:(A0Event *)event
{
    [self blockThreadWithEvent:event];
}


- (void)a0NewPipe:(A0Event *)event
{
    id nodeId = [event node];
    id newThreadId = [event objectForKey:@"OtherThread"];
    A0bThread *newThread;
    int protocol = [[event objectForKey:@"Protocol"] intValue];

    [event setObject:[NSString stringWithFormat:@"%d (%@%@%@%@)", protocol,
             (protocol & A0SyncPipe) ? @"Sync" : @"Async",
             (protocol & A0FlowPipe) ? @", Flow" : @"",
             (protocol & A0OutputPipe) ? @", Output" : @"",
             (protocol & A0InputPipe) ? @", Input" : @""] forKey:@"Protocol"];

    if (!(protocol & A0SyncPipe)) {
        // Assynchronous pipes create a thread to treat them; no join--detach
        newThread = [self startThread:newThreadId
                               atNode:nodeId
                               atTime:[event time] withEvent:event];
        [newThread setSystemThread:YES];
        [newThread detach];

        [self outputEntity:
            [A0Link newPipeLinkFromEvent:event toState:[newThread currentState]]];
    }

}
@end
