/* see A0bThread.h */

#import "A0bThread.h"
#import "A0bSimul.h"
#import "../General.bproj/UniqueString.h"
#import "../General.bproj/Macros.h"

PajeContainerType *A0ThreadEntityType;

@implementation A0bThread

+ (void)initialize
{
    if (self == [A0bThread class]) {
        A0ThreadEntityType = [[PajeContainerType alloc] initWithName:[UniqueString stringWithString:@"Thread"] containerType:[A0bNode entityType]];
    }
    [super initialize];
}

+ (PajeContainerType *)entityType { return A0ThreadEntityType; }

+ (A0bThread *)threadWithId:(id)thrId
                       node:(A0bNode *)n
                    service:(id)servId
               creationTime:(id)time
              creationEvent:(id)event
                  simulator:(A0bSimul *)simul
{
    return [[[self alloc] initWithId:thrId node:n service:servId
                        creationTime:time creationEvent:event
                           simulator:simul] autorelease];
}

- (id)initWithId:(id)thrId
            node:(A0bNode *)node
         service:(id)servId
    creationTime:(id)time
   creationEvent:(id)event
       simulator:(A0bSimul *)simul
{
//    int intThrId;
    self = [super initWithType:A0ThreadEntityType
                          name:thrId
                     container:node];
    if (!self)
        return nil;

    simulator = simul;
    serviceId = [servId retain];
    lastTime = nil;
    creationTime = [time retain];
    joinEntity = nil;
    status.detached = NO;
    status.terminated = NO;
    status.joined = NO;
    
//    intThrId = [localId intValue];
//    if ( (intThrId == 0) || (intThrId == 2) || (intThrId == 3) )
        status.system = YES;
//    else
//        status.system = NO;

//    NSLog(@"Thread %@ created by %@", [self description], [event description]);

        Assign(userEntities, [NSMutableDictionary dictionary]);

    // BUG: if thread is created with a0NewSlave, the event has @"OtherThread"
    // as the created thread, but it is not yet initialized to point to
    // the newly created thread (that is being created here), and
    // setCurrentState: output a state that doesn't know yet it's container!
    if ([[event name] isEqual:A0NewSlaveEventName])
        [event setObject:self forKey:@"OtherThread"];
    if ([[event name] isEqual:A0_AllocTraceBufferEventName])
        [event setThread:self];
    
    [self setCurrentState:[A0ThreadState runnableThreadStateFromEvent:event]];
    
    return self;
}

- (void)dealloc
{
    // doesn't release node, it has not been retained (see init..)
    [serviceId release];
    [currentState release];
    [lastTime release];
    [joinEntity release];
    [super dealloc];
}

- (void)setSystemThread:(BOOL)flag
{
    status.system = flag;
    if (currentState)
        [simulator outputEntity:currentState];
}

- (BOOL)isSystemThread
// return YES if thread is an Athapascan daemon:
//    0 = communication daemon
//    2 = remote thread creation daemon
//    3 = remote urgent thread creation daemon
{
    return status.system;
}

- (NSString *)printForDebugger
{
    return [NSString stringWithFormat:
        @"{Thread = %@; Node = %@; Service = %@; "
        @"LastTime = %@; "
        @"JoinEntity = %@; "
        @"Status = %s%s%s%s}",
        self, [self container], serviceId, lastTime, joinEntity,
        status.system?"System":"", status.detached?"Detached":"",
        status.joined?"Joined":"", status.terminated?"Terminated":""];
}

- (id)node { return container; }

- (void)terminateWithEvent:(id)event
                    atTime:(id)time
{
    if (status.terminated) {
        NSLog(@"Terminating already terminated thread %@ with event %@",
              self, event);
    } else {
        [self setLastTime:time];
        status.terminated = YES;
        
        [currentState setEndEvent:event];
//        [simulator outputEntity:currentState];
        [currentState release];
        currentState = nil;
        if ((!status.joined) && joinEntity) {
            [simulator outputEntity:[A0Link joinLinkFromEvent:event
                                                      toState:joinEntity]];
            Assign(joinEntity, nil);
            status.joined = YES;
        } else
            Assign(joinEntity, event);
    }
}

- (void)detach
{
    status.detached = YES;
}

- (BOOL)isDetached
{
    return status.detached;
}

- (BOOL)isTerminated
{
    return status.terminated;
}


// fix for a bug in the tracer (unblock events sometimes come after the
// first event of the thread after the block event)
- (void)setBugEvent:(A0Event *)event
{
    [event retain];
    [bugEvent release];
    bugEvent = event;
}

- (A0Event *)bugEvent
{
    return bugEvent;
}

- (BOOL)isWaitingForUnblockEvent
{
    return unblockTarget != nil;
}
// end of fix

- (int)intValue {
    return [[self name] intValue];
}

- (void)setLastTime:(id)time
{
    [lastTime release];
    lastTime = [time retain];
}
- (id)lastTime
{
    return lastTime;
}

- (NSDate *)startTime { return creationTime; }

- (NSDate *)endTime {
    if (lastTime)
        return lastTime;
    return [container endTime];
}

- (void)setJoinState:(id)state
{
    if (status.joined)
        NSLog(@"A0bSimul: state %@ joining the already joined thread %@",
               [state description], [self description]);
    else
        if (status.terminated) {
            [simulator outputEntity:
                [A0Link joinLinkFromEvent:joinEntity toState:state]];
            Assign(joinEntity, nil);
        } else
            Assign(joinEntity, state);
}

- (id)threadId
{
    return name;
}

- (id)serviceId
{
    return serviceId ? serviceId : @"Unknown";
}

- (A0ThreadState *)currentState
{
    return currentState;
}

- (void)setCurrentState:(A0ThreadState *)newCurrentState
{
    A0ThreadState *oldCurrentState = currentState;
    currentState = [newCurrentState retain];
    if (oldCurrentState) {
        if (newCurrentState) {
            [oldCurrentState setEndEvent:[newCurrentState startEvent]];
        }
        [oldCurrentState release];
    }
    if (currentState)
        [simulator outputEntity:currentState];
}



- (void)addRelatedEntity:(id)entity
{
    [currentState addRelatedEntity:entity];
}

- (void)blockWithEvent:(A0Event *)event
       unblockSelector:(SEL)selector target:(id)target
{
    unblockSelector = selector;
    unblockTarget = target;

    [self setCurrentState:[A0ThreadState blockedThreadStateFromEvent:event]];
}

- (void)a0UnblockThread:(A0Event *)event
{
    if (unblockTarget) {
        // tell the object where thread is blocked that it has unblocked
        // *** Must be done before changing currentState,
        // because unblockTarget can change relatedEntities.
        [unblockTarget performSelector:unblockSelector
                            withObject:self withObject:event];
        unblockTarget = nil;
    }

    [self setCurrentState:[A0ThreadState runnableThreadStateFromEvent:event]];
}


// NSCoding Protocol
- (void)encodeWithCoder:(NSCoder *)coder
{
    BOOL b[4] = {
        status.system,
        status.detached,
        status.terminated,
        status.joined,
    };
    [super encodeWithCoder:coder];
    [coder encodeObject:serviceId];
    [coder encodeObject:lastTime];
    [coder encodeObject:joinEntity];
    [coder encodeArrayOfObjCType:@encode(BOOL) count:4 at:&b];
}

- (id)initWithCoder:(NSCoder *)coder
{
    BOOL b[4];

    self = [super initWithCoder:coder];
    if (!self)
        return nil;
    
    serviceId = [[coder decodeObject] retain];
    lastTime = [[coder decodeObject] retain];
    joinEntity = [[coder decodeObject] retain];
    [coder decodeArrayOfObjCType:@encode(BOOL) count:4 at:&b];
    status.system = b[0];
    status.detached = b[1];
    status.terminated = b[2];
    status.joined = b[3];

    return self;
}
@end

#if 0
@implementation A0bThread (UserStates)
- (NSMutableArray *)userStateArrayForEntityType:(id)entityType
{
    NSMutableArray *userStateArray;
    userStateArray = [userStates objectForKey:entityType];
    if (!userStateArray) {
        userStateArray = [[NSMutableArray alloc] init];
        if (!userStates)
            Assign(userStates, [NSMutableDictionary dictionaryWithObject:userStateArray forKey:entityType]);
        else
            [userStates setObject:userStateArray forKey:entityType];
        [userStateArray release];
    }
    return userStateArray;
}

- (void)changeUserStateOfType:(id)entityType
                      toState:(id)name
                    withEvent:(A0Event*)event
{
    UserState *currentUserState, *newUserState;
    NSMutableArray *userStateArray;

    newUserState = [UserThreadState stateWithStartEvent:event
                                             entityType:entityType
                                             entityName:name];
    
    userStateArray = [self userStateArrayForEntityType:entityType];
    currentUserState = [userStateArray lastObject];
    if (currentUserState) {
        [currentUserState setEndEvent:event];
        [simulator entityChangedEndTime:currentUserState];
        [userStateArray replaceObjectAtIndex:[userStateArray count]-1
                                  withObject:newUserState];
    } else {
        [userStateArray addObject:newUserState];
    }
    
    [simulator outputEntity:newUserState];
}

- (void)pushUserStateOfType:(id)entityType
                    toState:(id)name
                  withEvent:(A0Event*)event
{
    UserState *newUserState;
    NSMutableArray *userStateArray;

    newUserState = [UserThreadState stateWithStartEvent:event
                                             entityType:entityType
                                             entityName:name];

    userStateArray = [self userStateArrayForEntityType:entityType];
    [userStateArray addObject:newUserState];

    [simulator outputEntity:newUserState];
}

- (void)popUserStateOfType:(id)entityType
                 withEvent:(A0Event *)event
{
    UserState *currentUserState;
    NSMutableArray *userStateArray;

    userStateArray = [self userStateArrayForEntityType:entityType];
    currentUserState = [userStateArray lastObject];
    if (currentUserState) {
        [currentUserState setEndEvent:event];
        [simulator entityChangedEndTime:currentUserState];
        [userStateArray removeLastObject];
    } else {
        NSLog(@"Thread has no user state to pop with event %@", event);
    }
}
@end
#endif
