/* DisplayFilter.m created by benhur on Sat 14-Jun-1997 */

#import "DisplayFilter.h"
#import "../General.bproj/FoundationAdditions.h"
#import "../General.bproj/UniqueString.h"

@implementation DisplayFilter

- (id)init
{
    NSWindow *win;
    
    [super init];

    filters = [[NSMutableDictionary alloc] init];

    if (![NSBundle loadNibNamed:@"DisplayFilter" owner:self]) {
        NSRunAlertPanel(@"DisplayFilter", @"Couldn't load interface file",
                        nil, nil, nil);
        return self;
    }

    win = [view window];
    
    // view is an NSBox. we need its contents.
    view = [view contentView];

    [matrix registerForDraggedTypes:[NSArray arrayWithObject:NSColorPboardType]];
    [matrix setDelegate:self];

    [entityTypePopUp removeAllItems];

    [[NSNotificationCenter defaultCenter] postNotificationName:@"PajeRegisterFilterNotification"
                      object:self
                    userInfo:[NSDictionary dictionaryWithObjectsAndKeys:
                                @"By type",  @"Name",
                                view,       @"View",
                                self,       @"Delegate", nil]];

    [[NSNotificationCenter defaultCenter]
        addObserver:self
           selector:@selector(filterEntityNameNotification:)
               name:@"PajeFilterEntityNameNotification"
             object:nil];

#ifndef GNUSTEP
    [win release];
#endif

    return self;
}

- (void)dealloc
{
    [filters release];
    [super dealloc];
}


//
// Notifications sent by other filters
//
- (void)hierarchyChanged
{
    MiscTree *hierarchy;
    NSEnumerator *hierEnum;
    
    hierEnum = [[inputComponent hierarchyTree] preorderAllSubtreeEnumerator];
    while (hierarchy = [hierEnum nextObject]) {
        if ([hierarchy isLeaf] && [hierarchy entityType])
            [self setFilterForEntityType:[hierarchy entityType]];
    }
    [self synchronizeMatrix];
    [super hierarchyChanged];
}

- (void)colorChangedForEntityType:(PajeEntityType *)entityType
{
    if (!entityType || [entityType isEqual:[self selectedEntityType]])
        [self synchronizeMatrix];
    [super colorChangedForEntityType:entityType];
}



//
//
//
- (void)setFilterForEntityType:(PajeEntityType *)entityType
// a new entity type is known. create a new filter for it.
{
    NSMutableSet *filter;

    NSParameterAssert(entityType);
    
    filter = [filters objectForKey:entityType];

    if (nil == filter) {

        filter = [self defaultFilterForEntityType:entityType];

        [filters setObject:filter forKey:entityType];

        [entityTypePopUp addItemWithTitle:[entityType description]];
        [[entityTypePopUp lastItem] setRepresentedObject:entityType];
#ifdef GNUSTEP
	[entityTypePopUp setEnabled:YES];
#endif
    }
}

- (NSMutableSet *)defaultFilterForEntityType:(PajeEntityType *)entityType
{
    NSString *defaultName = [NSString stringWithFormat:@"%@%@Filter", NSStringFromClass([self class]), [entityType description]];
    NSArray *filterArray = [[NSUserDefaults standardUserDefaults] arrayForKey:defaultName];
    NSMutableSet *uniqueSet;
    NSEnumerator *filterEnum;
    NSString *s;
    
#ifdef GNUSTEP
    uniqueSet = [[[NSMutableSet alloc] init] autorelease];
#else
    uniqueSet = [NSMutableSet set];
#endif
    filterEnum = [filterArray objectEnumerator];
    while (s = [filterEnum nextObject])
        [uniqueSet addObject:[UniqueString stringWithString:s]];
    
    return uniqueSet;
}

- (void)registerDefaultFilter:(NSSet *)filter
                forEntityType:(PajeEntityType *)entityType
{
    NSString *defaultName = [NSString stringWithFormat:@"%@%@Filter", NSStringFromClass([self class]), [entityType description]];
    NSArray *filterArray = [filter allObjects];
    [[NSUserDefaults standardUserDefaults] setObject:filterArray
                                              forKey:defaultName];
}

- (PajeEntityType *)selectedEntityType
{
    return [[entityTypePopUp selectedItem] representedObject];
}


//
// Handle interface messages
//
- (void)selectAll:(id)sender
{
    PajeEntityType *entityType = [self selectedEntityType];
    NSMutableSet *filter = [filters objectForKey:entityType];

    if (filter) {
        [filter removeAllObjects];
        [self synchronizeMatrix];
        [self registerDefaultFilter:filter forEntityType:entityType];
        [super dataChangedForEntityType:entityType];
    } else {
        [NSException raise:@"DisplayFilter: Internal Inconsistency"
                    format:@"%@ sent by unknown view %@", NSStringFromSelector(_cmd), sender];
    }
}

- (void)unselectAll:(id)sender
{
    PajeEntityType *entityType = [self selectedEntityType];
    NSMutableSet *filter = [filters objectForKey:entityType];

    if (filter) {
        [filter addObjectsFromArray:[inputComponent allNamesForEntityType:entityType]];
        [self synchronizeMatrix];
        [self registerDefaultFilter:filter forEntityType:entityType];
        [super dataChangedForEntityType:entityType];
    } else {
        [NSException raise:@"DisplayFilter: Internal Inconsistency"
                    format:@"%@ sent by unknown view %@", NSStringFromSelector(_cmd), sender];
    }
}

- (void)filterName:(NSString *)name
      ofEntityType:(PajeEntityType *)entityType
              show:(BOOL)flag
{
    NSMutableSet *filter;

    filter = [filters objectForKey:entityType];

    if (filter != nil) {
        if (flag)
            [filter removeObject:name];
        else
            [filter addObject:name];
        [self registerDefaultFilter:filter forEntityType:entityType];
        [super dataChangedForEntityType:entityType];
    }
}

- (void)filterEntityNameNotification:(NSNotification *)notification
{
    NSString *name;
    PajeEntityType *entityType;
    BOOL flag;
    NSDictionary *userInfo;

    userInfo = [notification userInfo];
    if (userInfo == nil)
        return;

    entityType = [userInfo objectForKey:@"EntityType"];
    name = [userInfo objectForKey:@"EntityName"];
    flag = [[userInfo objectForKey:@"Show"] isEqualToString:@"YES"];

    if (entityType != nil && name != nil)
        [self filterName:name
            ofEntityType:entityType
                    show:flag];
    [self synchronizeMatrix];        
}

- (void)matrixChanged:(id)sender
{
    NSButtonCell *cell = [sender selectedCell];
    PajeEntityType* entityType = [self selectedEntityType];

    if (cell != nil) {
        [self filterName:[cell representedObject]
            ofEntityType:entityType
                    show:[cell state]];
    }
}

- (void)entityPopUpChanged:(id)sender
{
    [self synchronizeMatrix];
}

- (void)synchronizeMatrix
{
    id entityType = [self selectedEntityType];
    NSMutableSet *filter;
    NSEnumerator *names;
    id entityName;
    int i = 0;
    NSButtonCell *cell;
    NSMutableAttributedString *title;

    if (!entityType)
        return;
    
    filter = [filters objectForKey:entityType];
    names = [[inputComponent allNamesForEntityType:entityType] objectEnumerator];

    while ([matrix cellAtRow:0 column:0])
        [matrix removeRow:0];

    while (entityName = [names nextObject]) {
        [matrix addRow];
        cell = [matrix cellAtRow:i column:0];
        [cell setState:!([filter containsObject:entityName])];
        [cell setRepresentedObject:entityName];
#ifndef GNUSTEP
        title = [[[NSMutableAttributedString alloc] initWithString:[entityName description]] autorelease];
#if defined(__APPLE__)
        [title replaceCharactersInRange:NSMakeRange(0,0) withString:@"* "];
        [title addAttribute:NSFontAttributeName
                      value:[NSFont boldSystemFontOfSize:18]
                      range:NSMakeRange(0,1)];
        [title addAttribute:NSFontAttributeName
                      value:[NSFont systemFontOfSize:12]
                      range:NSMakeRange(1, [title length]-1)];
#else
        // 0x25a0 is unicode for a solid small square
        [title replaceCharactersInRange:NSMakeRange(0,0) withString:[NSString stringWithCharacter:'#'/*0x25a0*/]];
        [title replaceCharactersInRange:NSMakeRange(1,0) withString:@" "];
#endif
        [title addAttribute:NSForegroundColorAttributeName
                      value:[inputComponent colorForName:entityName
                                            ofEntityType:entityType]
                      range:NSMakeRange(0,1)];
        [title fixAttributesInRange:NSMakeRange(0, [title length])];
        [cell setAttributedTitle:title];
#else /*!GNUSTEP*/
        [cell setTitle:[entityName description]];
#endif
        [cell setHighlightsBy:NSChangeBackgroundCellMask];
        i++;
    }
    [matrix sortUsingSelector:@selector(compare:)];
    [matrix sizeToCells];
    [matrix setNeedsDisplay:YES];
}

//
// interact with interface
//

- (void)viewWillBeSelected:(NSView *)selectedView
// message sent by PajeController when a view will be selected for display
{
    [self synchronizeMatrix];        
}


//
// Trace messages that are filtered
//
- (NSEnumerator *)enumeratorOfEntityType:(PajeEntityType *)entityType
                             inContainer:(PajeContainer *)container
                                fromTime:(NSDate *)start
                                  toTime:(NSDate *)end;
{
    NSEnumerator *origEnum;
    NSMutableSet *filter = [filters objectForKey:entityType];

    origEnum = [inputComponent enumeratorOfEntityType:entityType
                                          inContainer:container
                                             fromTime:start
                                               toTime:end];

    if (filter) {
        return [[[FilteredEnumerator alloc] initWithEnumerator:origEnum filter:filter selector:@selector(name)] autorelease];
    } else {
        return origEnum;
    }
}

- (NSArray *)allNamesForEntityType:(PajeEntityType *)entityType
{
    NSArray *allNames = [inputComponent allNamesForEntityType:entityType];
    NSMutableSet *filter = [filters objectForKey:entityType];

    if (filter) {
        NSMutableSet *set = [NSMutableSet setWithArray:allNames];
        [set minusSet:filter];
        return [set allObjects];
    } else
        return allNames;
}


//
// methods to be a dragging destination, as delegate of matrix
//

- (void)highlightCell:(id)cell inMatrix:(NSMatrix *)m
{
    int r, c;
    static id highlightedCell; // the last cell that has been highlighted
    
    if (highlightedCell && ![highlightedCell isEqual:cell]) {
        [m getRow:&r column:&c ofCell:highlightedCell];
        [m highlightCell:NO atRow:r column:c];
        highlightedCell = nil;
    }
    
    if (cell && ![cell isEqual:highlightedCell]) {
        [m getRow:&r column:&c ofCell:cell];
        [m highlightCell:YES atRow:r column:c];
        highlightedCell = cell;
    }
}

- (unsigned int)matrix:(NSMatrix *)m draggingUpdated:(id <NSDraggingInfo>)sender
{
    NSPoint point = [m convertPoint:[sender draggingLocation] fromView:nil];
    NSButtonCell *cell = [m cellAtPoint:point];

    [self highlightCell:cell inMatrix:m];
    if (cell)
        return NSDragOperationGeneric;
    else
        return NSDragOperationNone;
}

#ifdef GNUSTEP
- (BOOL)matrix:(NSMatrix *)m prepareForDragOperation:(id <NSDraggingInfo>)sender
{
    return YES;
}
#endif

- (BOOL)matrix:(NSMatrix *)m performDragOperation:(id <NSDraggingInfo>)sender
{
    NSPoint point;
    NSButtonCell *cell;

    point = [m convertPoint:[sender draggingLocation] fromView:nil];
    cell = [m cellAtPoint:point];

    [self highlightCell:nil inMatrix:m];
    
    if (cell) {
        id entityType = [self selectedEntityType];
        [inputComponent setColor:[NSColor colorFromPasteboard:[sender draggingPasteboard]]
                         forName:[cell representedObject]
                    ofEntityType:entityType];
        if ([cell state] == 1)
            [super colorChangedForEntityType:entityType];
        [self synchronizeMatrix];
        return YES;
    }
    return NO;
}

- (void)matrix:(NSMatrix *)m draggingExited:(id <NSDraggingInfo>)sender
{
    [self highlightCell:nil inMatrix:m];
}

@end
