  /*
   * Encapsulate.m
   *
   * Component that reads individual events, states and communications
   * and encapsulates them for grouped access.
   *
   * 19970423 BS  creation
   */

#import "Encapsulate.h"
#import "../General.bproj/TimeList.h"
#import "../General.bproj/MultiEnumerator.h"
#import "../General.bproj/PajeEntity.h"

@implementation Encapsulate

- (id)init
{
    [super init];
    
    traceChanged = NO;
    timerActive = NO;

    entityLists = [[NSMutableDictionary alloc] init];
    entityTypeToInstances = [[NSMutableDictionary alloc] init];
    entityTypeToClass = [[NSMutableDictionary alloc] init];
    startTime = nil;
    endTime = nil;
    startTimeInMemory = nil;
    endTimeInMemory = nil;
    hierarchyTree = [[MiscTree alloc] init];

    [[NSNotificationCenter defaultCenter]
        addObserver:self
           selector:@selector(fileSelected:)
               name:@"PajeFilenameNotification"
             object:nil];
    [[NSNotificationCenter defaultCenter]
        addObserver:self
           selector:@selector(containerSelected:)
               name:@"PajeSelectContainerNotification"
             object:nil];

    return self;
}

- (void)dealloc
{
    [entityLists release];
    [entityTypeToInstances release];
    [entityTypeToClass release];
    [startTime release];
    [endTime release];
    [startTimeInMemory release];
    [endTimeInMemory release];
    [hierarchyTree release];

    [super dealloc];
}

- (void)reset
/*"Forget everything that has been done, start all over again."*/
{
    traceChanged = YES;
    if (!timerActive) {
        [self performSelector:@selector(timeElapsed:) withObject:self afterDelay:0];
        timerActive = YES;
    }
}

- (void)fileSelected:(id)notification
/*"notification sent by the controller when a new file is selected."*/
{
    NSDictionary *userInfo = [notification userInfo];

    startTime = [[userInfo objectForKey:@"StartTime"] retain];
    endTime = [[userInfo objectForKey:@"EndTime"] retain];
    startTimeInMemory = [startTime retain];
    endTimeInMemory = [startTime retain];
//    [self reset];
}

- (void)containerSelected:(id)notification
/*"notification sent when a container is selected."*/
{
    id container;
    NSDictionary *userInfo = [notification userInfo];

    container = [userInfo objectForKey:@"Container"];
    if (container != nil && [container isKindOfClass:[PajeContainer class]])
        [self filterContainer:container show:NO];
}

- (void)inputEntity:(id)entity
/*"There's new data in the input port. Encapsulate it."*/
{
    [self addEntity:entity];
    traceChanged = YES;

    if (!timerActive) {
        [self performSelector:@selector(timeElapsed:)
                   withObject:self
                   afterDelay:0.0];
        timerActive = YES;
    }
}


- (void)addEntity:(id<PajeEntity,NSObject>)entity
{
    PajeEntityType *entityType = [entity entityType];
    NSMutableDictionary *dict = [entityLists objectForKey:entityType];
    id container = [entity container];
    TimeList *list;

    //KLUDGE to identify last date in memory.
    // (doesn't work if events are not seen)
    // FIXME this should be calculated on demand
    if ([entity drawingType] == PajeEventDrawingType) {
        NSDate *time = [entity time];
        if([time compare:endTimeInMemory] == NSOrderedDescending) {
            [endTimeInMemory release];
            endTimeInMemory = [time retain];
        }
    }
    //ENDKLUDGE

    if (!dict) {
        // unknown entity type -- create a new dict for it; add to hierarchy
        dict = [[NSMutableDictionary alloc] init];
        [entityLists setObject:dict forKey:entityType];
        [dict release];

        [self _addToHierarchyTree:entity];

        [entityTypeToClass setObject:[entity class] forKey:entityType];
    }

    list = [dict objectForKey:container];

    if (!list) {
        // unknown container for this entity type -- create a new list for it
        list = [[TimeList alloc] init];
        [dict setObject:list forKey:container];
        [list release]; // dict is retaining it

        {
            // Generate entityTypeToInstances for all containers
            // FIXME this info should be in hierarchyTree
            id instance = entity;
            while ((container != nil) && (container != instance)) {
                id entityType = [container entityType];
                NSMutableSet *set;

                set = [entityTypeToInstances objectForKey:entityType];
                if (!set) {
                    set = [NSMutableSet setWithObject:container];
                    [entityTypeToInstances setObject:set forKey:entityType];
                } else {
                    [set addObject:container];
                }
                instance = container;
                container = [instance container];
            }
        }
    }

    [list addObject:entity];
}

- (void)_addToHierarchyTree:(id)entity
{
    // FIXME there should be a PajeHierarchy class to ease this
    NSMutableArray *pathToRoot;
    NSEnumerator *types;
    id entityType;
    id root;
    id ent;
    MiscTree *tree;

    // take hierarchy of containers of entity up to root
    pathToRoot = [NSMutableArray array];
    ent = entity;
    root = nil;
    while (ent != root) {
        [pathToRoot addObject:[ent entityType]];
        root = ent;
        ent = [ent container];
    }

    // set root of tree if not already set
    tree = [self hierarchyTree];
    if ([tree entityType] == nil) {
        rootInstance = root;
        [tree setEntityType:[root entityType]];
    } else
        if (![[tree entityType] isEqual:[root entityType]])
            NSLog(@"Inconsistency in hierarchy tree: %@ != %@",
                  [tree entityType], [root entityType]);

    // insert all entity types in tree
    types = [pathToRoot reverseObjectEnumerator];
    while (entityType = [types nextObject]) {
        NSEnumerator *subtreeEnum = [tree subtreeEnumerator];
        MiscTree *subtree;

        while (subtree = [subtreeEnum nextObject]) {
            if ([[subtree entityType] isEqual:entityType])
                break;
        }
        if (subtree == nil) {
            subtree = [MiscTree tree];
            [subtree setEntityType:entityType];
            [tree addSubtree:subtree];
        }
        tree = subtree;
    }
}


- (void)entityChangedEndTime:(id<PajeEntity>)entity
{
    PajeEntityType *entityType = [entity entityType];
    NSMutableDictionary *dict = [entityLists objectForKey:entityType];
    id container = [entity container];
    TimeList *list;

    list = [dict objectForKey:container];

    if (!list) {
        NSLog(@"entity %@ not found (entityChangedEndTime)");
        return;
    }

    [list objectChangedEndTime:entity];
}


- (void)send
/*"notifies that data has changed."*/
{
    // only send trace if it is already initialized
    // (or else space time diagram will think trace has no data, and the
    // lazy reading will not work well)
    if (startTime) {
        // FIXME should send more specific notification
        [self hierarchyChanged];
        traceChanged = NO;
        [self performSelector:@selector(timeElapsed:) withObject:self afterDelay:0.5];
        timerActive = YES;
    }
}

- (void)timeElapsed:(id)sender
/*"Called by the timer; notifies if necessary."*/
{
    timerActive = NO;
    if (traceChanged) {
        [self send];
    }
}


- (NSDate *)time { return startTime; }

// time when the trace starts
- (NSDate *)startTime { return startTime; }

// time when the trace ends
- (NSDate *)endTime { return endTime; }

- (NSDate *)firstTime { return startTime; }
- (NSDate *)lastTime { return endTime; }

- (NSDate *)startTimeInMemory { return startTimeInMemory; }
- (NSDate *)endTimeInMemory { return endTimeInMemory; }

- (MiscTree *)hierarchyTree { return hierarchyTree; }

- (id)rootInstance
{
    return rootInstance;
}

- (void)verifyStartTime:(NSDate *)start endTime:(NSDate *)end
{
    NSDictionary *userInfo = nil;
    if ([start earlierThanDate:startTimeInMemory]
        && [startTimeInMemory laterThanDate:startTime]) {
        userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
            [start laterDate:startTime], @"StartTime",
            [end earlierDate:endTime], @"EndTime",
            nil];
    } else if ([end laterThanDate:endTimeInMemory]
               && [endTimeInMemory earlierThanDate:endTime]) {
        userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
            [start laterDate:startTime], @"StartTime",
            [end earlierDate:endTime], @"EndTime",
            nil];
    }
    if (userInfo) {
//        NSLog(@"Trace notificando:%@ p(%@ %@) m(%@ %@)", userInfo, [start description], [end description], [startTimeInMemory description], [endTimeInMemory description]);
        [[NSNotificationCenter defaultCenter]
                      postNotificationName:@"PajeTraceNotInMemoryNotification"
                                    object:self
                                  userInfo:userInfo];
//        NSLog(@"fim da notificacao");
    }
}

- (NSEnumerator *)enumeratorOfEntityType:(PajeEntityType *)entityType
                             inContainer:(PajeContainer *)container
                                fromTime:(NSDate *)start
                                  toTime:(NSDate *)end
{
    NSDictionary *dict;
    TimeList *list;
    NSEnumerator *subcontainerEnum;
    id subcontainer;
    MultiEnumerator *multiEnum;
    
    [self verifyStartTime:start endTime:end];
    
    dict = [entityLists objectForKey:entityType];
    // if container directly contains entityType's
    list = [dict objectForKey:container];
    if (list)
        return [list timeEnumeratorFromTime:start toTime:end];

    // if container should contain entityType directly, we have no entities!
    if ([[entityType containerType] isEqual:[container entityType]])
        return nil;
    
    // container does not contain entityType's directly. Try containers in-between.
    subcontainerEnum = [self enumeratorOfEntityType:[entityType containerType]
                                        inContainer:container];
    multiEnum = [MultiEnumerator enumerator];
    while (subcontainer = [subcontainerEnum nextObject]) {
        list = [dict objectForKey:subcontainer];
        if (list)
            [multiEnum addEnumerator:[list timeEnumeratorFromTime:start
                                                           toTime:end]];
    }
    return multiEnum;
}


- (NSEnumerator *)enumeratorOfEntityType:(PajeEntityType *)entityType
                             inContainer:(PajeContainer *)container
{
    NSEnumerator *ienum;
    id instance;
    NSMutableArray *instancesInContainer = [NSMutableArray array];

    ienum = [[entityTypeToInstances objectForKey:entityType] objectEnumerator];
    while (instance = [ienum nextObject]) {
        if ([instance isContainedBy:container]) {
            [instancesInContainer addObject:instance];
        }
    }
    return [[instancesInContainer sortedArrayUsingSelector:@selector(compare:)]
objectEnumerator];
}

- (NSNumber *)minValueForEntityType:(PajeEntityType *)entityType
{
    if ([entityType respondsToSelector:@selector(minValue)])
        return [(PajeVariableType *)entityType minValue];
    else
        return [NSNumber numberWithInt:0];
}

- (NSNumber *)maxValueForEntityType:(PajeEntityType *)entityType
{
    if ([entityType respondsToSelector:@selector(maxValue)])
        return [(PajeVariableType *)entityType maxValue];
    else
        return [NSNumber numberWithInt:0];
}

- (NSNumber *)minValueForEntityType:(PajeEntityType *)entityType
                        inContainer:(PajeContainer *)container
{
    if ([container respondsToSelector:@selector(minValueForEntityType:)])
        return [container minValueForEntityType:entityType];
    else
        return [NSNumber numberWithInt:0];
}

- (NSNumber *)maxValueForEntityType:(PajeEntityType *)entityType
                        inContainer:(PajeContainer *)container
{
    if ([container respondsToSelector:@selector(maxValueForEntityType:)])
        return [container maxValueForEntityType:entityType];
    else
        return [NSNumber numberWithInt:0];
}

- (NSNumber *)valueForEntity:(PajeEntity *)entity
{
    if ([entity respondsToSelector:@selector(value)])
        return [entity value];
    else
        return [NSNumber numberWithInt:0];
}

- (PajeContainer *)containerForEntity:(id<PajeEntity>)entity;
{
    return [entity container];
}
- (PajeEntityType *)entityTypeForEntity:(id<PajeEntity>)entity;
{
    return [entity entityType];
}
- (PajeContainer *)sourceContainerForEntity:(id<PajeLink>)entity;
{
    return [entity sourceContainer];
}
- (PajeEntityType *)sourceEntityTypeForEntity:(id<PajeLink>)entity;
{
    return [entity sourceEntityType];
}
- (PajeContainer *)destContainerForEntity:(id<PajeLink>)entity;
{
    return [entity destContainer];
}
- (PajeEntityType *)destEntityTypeForEntity:(id<PajeLink>)entity;
{
    return [entity destEntityType];
}

- (NSArray *)relatedEntitiesForEntity:(id<PajeEntity>)entity
{
    return [entity relatedEntities];
}

- (NSColor *)colorForEntity:(id<PajeEntity>)entity
{
    return [entity color];
}

- (NSDate *)startTimeForEntity:(id<PajeEntity>)entity;
{
    return [entity startTime];
}

- (NSDate *)endTimeForEntity:(id<PajeEntity>)entity;
{
    return [entity endTime];
}

- (NSDate *)timeForEntity:(id<PajeEntity>)entity
{
    return [entity time];
}

- (PajeDrawingType)drawingTypeForEntity:(id<PajeEntity>)entity
{
    return [entity drawingType];
}

- (NSString *)nameForEntity:(id<PajeEntity>)entity
{
    return [entity name];
}

- (int)imbricationLevelForEntity:(id<PajeEntity>)entity
{
    return [entity imbricationLevel];
}

- (NSString *)descriptionForEntity:(id<PajeEntity>)entity
{
    return [NSString stringWithFormat:@"%@ [%@ in %@ %@]",
        [entity name], [entity entityType],
        [[entity container] entityType], [entity container]];
}

- (BOOL)canHighlightEntity:(id<PajeEntity>)entity
{
    return YES;
}



- (NSArray *)allNamesForEntityType:(PajeEntityType *)entityType
{
    if ([entityType respondsToSelector:@selector(allNames)])
        return [entityType allNames];
    return [[entityTypeToClass objectForKey:entityType] allNames];
}

- (PajeDrawingType)drawingTypeForEntityType:(PajeEntityType *)entityType
{
    Class entityClass;

    if ([entityType respondsToSelector:@selector(drawingType)])
        return [entityType drawingType];
    entityClass = [entityTypeToClass objectForKey:entityType];
    if ([entityClass respondsToSelector:@selector(drawingType)])
        return [entityClass drawingType];
    else
        return PajeStateDrawingType;
}

- (NSColor *)colorForName:(NSString *)name
             ofEntityType:(PajeEntityType *)entityType
{
    if ([entityType respondsToSelector:@selector(colorForName:)])
        return [entityType colorForName:name];
    return [[entityTypeToClass objectForKey:entityType] colorForName:name];
}

- (void)setColor:(NSColor *)color
         forName:(NSString *)name
    ofEntityType:(PajeEntityType *)entityType
{
    if ([entityType respondsToSelector:@selector(setColor:forName:)])
        [entityType setColor:color forName:name];
    else
        [[entityTypeToClass objectForKey:entityType] setColor:color
                                                      forName:name];
}

- (void)setColor:(NSColor *)color forEntity:(id<PajeEntity>)entity
{
    [entity setColor:color];
}

- (NSColor *)colorForEntityType:(PajeEntityType *)entityType
{
    if ([entityType respondsToSelector:@selector(color)])
        return [entityType color];
    return [NSColor blackColor];
}



- (void)removeObjectsBeforeTime:(NSDate *)time
{
    NSEnumerator *e = [entityLists objectEnumerator];
    NSMutableDictionary *d;

    if ([time earlierDate:startTimeInMemory] == time)
        return;

//    NSLog(@"Trace removing before %@", [time description]);
    while (d = [e nextObject])
#if 0
        //xxx waiting for NSDictionary to implement makeObjectsPerformSelector:
        [d makeObjectsPerformSelector:@selector(removeObjectsBeforeTime:) withOb
ject:time];
#else
    {
        NSEnumerator *e2 = [d objectEnumerator];
        TimeList *l;
        while (l = [e2 nextObject])
            [l removeObjectsBeforeTime:time];
    }
#endif

    [time retain];
    [startTimeInMemory release];
    startTimeInMemory = time;

    if ([time laterDate:endTimeInMemory] == time) {
        [endTimeInMemory release];
        endTimeInMemory = [time retain];
    }
}

- (void)removeObjectsAfterTime:(NSDate *)time
{
    NSEnumerator *e = [entityLists objectEnumerator];
    NSMutableDictionary *d;

    if ([time laterDate:endTimeInMemory] == time)
        return;

//    NSLog(@"Trace removing after %@", [time description]);
    while (d = [e nextObject])
#if 0
        //xxx waiting for NSDictionary to implement makeObjectsPerformSelector:
        [d makeObjectsPerformSelector:@selector(removeObjectsAfterTime:) withObj
ect:time];
#else
    {
        NSEnumerator *e2 = [d objectEnumerator];
        TimeList *l;
        while (l = [e2 nextObject])
            [l removeObjectsAfterTime:time];
    }
#endif

    [time retain];
    [endTimeInMemory release];
    endTimeInMemory = time;

    if ([time earlierDate:startTimeInMemory]) {
        [startTimeInMemory release];
        startTimeInMemory = [time retain];
    }
}

- (void)inspectEntity:(id<PajeInspecting>)entity
{
    [entity inspect];
}

@end
