// DrawView+Positioning
// --------------
// methods for finding positions of entities in DrawView.

#import "DrawView.h"

@implementation DrawView (Positioning)

- (void)setRect:(NSRect)rect
  forEntityType:(PajeEntityType *)type
    inContainer:(PajeContainer *)container
{
    STEntityDescriptor *entityDescriptor;

    entityDescriptor = [self descriptorForEntityType:type];

//    NSLog(@"%5d%5d for %@ in %@", (int)rect.origin.y, (int)rect.size.height, type, container);
    if (entityDescriptor) {
        [entityDescriptor setRect:rect forContainer:container];
    }
}

- (NSRect)rectForEntityType:(PajeEntityType *)type
                inContainer:(PajeContainer *)container
{
    STEntityDescriptor *entityDescriptor;

    entityDescriptor = [self descriptorForEntityType:type];

    if (entityDescriptor) {
        return [entityDescriptor rectInContainer:container];
    }
    return NSZeroRect;
}

- (NSRect)rectForInstance:(id)entity
             ofEntityType:(PajeEntityType *)type
{
    STEntityDescriptor *entityDescriptor;

    entityDescriptor = [self descriptorForEntityType:type];

    if (entityDescriptor) {
        return [entityDescriptor rectForInstance:entity];
    }
    return NSZeroRect;
}

- (NSRect)calcRectForEntityType:(PajeEntityType *)entityType
                    inContainer:(PajeContainer *)container
                           minY:(float)minY
{
    STEntityDescriptor *edesc = [self descriptorForEntityType:entityType];
    NSRect rect = [edesc drawingRect];
    if ([edesc drawingType] != PajeLinkDrawingType) {
        if ([edesc canDisplay]
            /* && [filter existEntityType:entityType
                              inContainer:container]*/) {
            rect.origin.y += minY;
        } else {
            rect.origin.y = minY;
            rect.size.height = 0;
        }
    }
    [self setRect:rect forEntityType:entityType inContainer:container];
    return rect;
}

- (NSRect)calcRectForInstance:(id)entity
              ofHierarchyTree:(MiscTree *)hierarchyTree
                         minY:(float)minY
{
    NSEnumerator *subtreeEnum;
    MiscTree *subtree;
    id entityType = [hierarchyTree entityType];
    STEntityDescriptor *edesc = [self descriptorForEntityType:entityType];
    NSRect rect = NSZeroRect;

    /*if ([hierarchyTree isLeaf])
        return [self calcRectForEntityType:entityType
                               inContainer:entity
                                      minY:minY];*/
    NSAssert(![hierarchyTree isLeaf], @"hierarchyTree shouldn't be leaf");

    subtreeEnum = [hierarchyTree subtreeEnumerator];
    while (subtree = [subtreeEnum nextObject]) {
        NSRect r;
        r = [self calcRectForAllInstancesOfHierarchyTree:subtree
                                             inContainer:entity
                                                    minY:minY];
        if (!NSIsEmptyRect(r)) {
            rect = NSUnionRect(rect, r);
            minY = NSMaxY(rect);
        }
    }
    [edesc setRect:rect forInstance:entity];
    return rect;
}

- (NSRect)calcRectForAllInstancesOfHierarchyTree:(MiscTree *)hierarchyTree
                                     inContainer:(PajeContainer *)container
                                            minY:(float)minY
{
    NSRect rect = NSZeroRect;
    id entityType = [hierarchyTree entityType];
    if ([hierarchyTree isLeaf]) {
        rect = [self calcRectForEntityType:entityType
                               inContainer:container
                                      minY:minY];
        if ([container respondsToSelector:@selector(startTime)]) {
            rect.origin.x += TIMEtoX([container startTime]);
            rect.size.width += TIMEtoX([endTime earlierDate:[container endTime]]) - TIMEtoX([container startTime]);
        } else {
            rect.origin.x += TIMEtoX(startTime);
            rect.size.width += TIMEtoX(endTime) - TIMEtoX(startTime);
        }
    } else {
        STEntityDescriptor *edesc = [self descriptorForEntityType:entityType];
        NSEnumerator *ienum;
        id instance;

        rect.origin.y = minY;
        rect.size.height = [edesc offset];
        if ([container respondsToSelector:@selector(startTime)]) {
            rect.origin.x = TIMEtoX([startTime laterDate:[container startTime]]);
            rect.size.width = TIMEtoX([endTime earlierDate:[container endTime]]) - TIMEtoX([startTime laterDate:[container startTime]]);
        }
        minY = NSMaxY(rect);
        
        // check all instances on this hierarchy
        ienum = [filter enumeratorOfEntityType:entityType
                                   inContainer:container];
        while (instance = [ienum nextObject]) {
            NSRect r;
            r = [self calcRectForInstance:instance
                          ofHierarchyTree:hierarchyTree
                                     minY:minY];
            if (!NSIsEmptyRect(r)) {
                rect = NSUnionRect(rect, r);
                rect.size.height += [edesc separation];
                minY = NSMaxY(rect);
            }
        }
        if (container)
            [edesc setRect:rect forContainer:container];
    }
    return rect;
}


- (NSArray *)verticalRulerDescriptors
{
    NSMutableArray *descriptors = [NSMutableArray array];
    NSEnumerator *l2enum = [[[filter hierarchyTree] subtrees] objectEnumerator];
    MiscTree *l2subtree;
    STEntityDescriptor *edesc;

    while (l2subtree = [l2enum nextObject]) {
        if (![l2subtree isLeaf]) {
            edesc = [self descriptorForEntityType:[l2subtree entityType]];
            if (edesc)
                [descriptors addObject:edesc];
            {
                NSEnumerator *l3enum = [[l2subtree subtrees] objectEnumerator];
                MiscTree *l3subtree;
                STEntityDescriptor *e3desc;

                while (l3subtree = [l3enum nextObject]) {
                    if (![l3subtree isLeaf]) {
                        e3desc = [self descriptorForEntityType:[l3subtree entityType]];
                        if (e3desc)
                            [descriptors addObject:e3desc];
                        {
                            NSEnumerator *l4enum = [[l3subtree subtrees] objectEnumerator];
                            MiscTree *l4subtree;
                            STEntityDescriptor *e4desc;

                            while (l4subtree = [l4enum nextObject]) {
                                if (![l4subtree isLeaf]) {
                                    e4desc = [self descriptorForEntityType:[l4subtree entityType]];
                                    if (e4desc)
                                        [descriptors addObject:e4desc];
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    
    return descriptors;
}

@end