/*
 */

#import "StatViewer.h"

#import "../A0bSimul.bproj/A0Comm.h"
#import "../General.bproj/Macros.h"

@implementation StatViewer
- init
{
    [super init];

    if (![NSBundle loadNibNamed:@"StatViewer" owner:self])
        NSRunAlertPanel(@"StatViewer", @"Couldn't load interface file",
                        nil, nil, nil);

    return self;
}

- (void)awakeFromNib
{
    id cell;
    NSSize cellSize = NSZeroSize;
    NSString *cellSizeString;

    [window setFrameAutosaveName:NSStringFromClass([self class])];

    cellSizeString = [[NSUserDefaults standardUserDefaults] stringForKey:@"StatViewerCellSize"];
    if (cellSizeString)
        cellSize = NSSizeFromString(cellSizeString);
    if (!NSEqualSizes(cellSize, NSZeroSize))
        [matrix setCellSize:cellSize];
    
    cell = [[PieCell alloc] initImageCell:nil];
    [cell setRepresentedObject:nil];
    [cell setDataProvider:self];
    [matrix setPrototype:cell];
    [matrix putCell:cell atRow:0 column:0];
    
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(timeSelectionChanged:) name:@"PajeTimeSelectionChangedNotification" object:nil];

    // notify that we are a Tool
    [[NSNotificationCenter defaultCenter]
              postNotificationName:@"PajeRegisterToolNotification"
                            object:self
                          userInfo:[NSDictionary dictionaryWithObjectsAndKeys:
                              self, @"Tool",
                              @"Statistics", @"Name",
                              nil]];
}

- (void)activateTool:(id)sender
{
    [window makeKeyAndOrderFront:self];
}

- (void)setInputComponent:(PajeComponent *)component
{
    [super setInputComponent:component];
    Assign(filter, component);
}

- (void)hierarchyChanged
{
    [self invalidateValues];
}

- (void)setStartTime:(NSDate *)time
{
    Assign(startTime, time);
}

- (void)setEndTime:(NSDate *)time
{
    Assign(endTime, time);
}

- (void)timeSelectionChanged:(NSNotification *)notification
{
#ifdef GNUSTEP
    NSLog(@"StatView recebeu notificacao: %@", notification);
#endif
    [self setStartTime:[[notification userInfo] objectForKey:@"StartTime"]];
    [self setEndTime:[[notification userInfo] objectForKey:@"EndTime"]];
    [durationField setStringValue:[NSString stringWithFormat:@"%.6f s", [endTime timeIntervalSinceDate:startTime]]];
    [self invalidateValues];
}

- (void)invalidateValues
{
    int i=0;
    NSEnumerator *nn;
    id container;
    PieCell *cell;

    if (!filter) return;

    nn = [filter enumeratorOfEntityType:@"Node" inContainer:[filter rootInstance]];
    while (container = [nn nextObject]) {
        [matrix renewRows:i+1 columns:1];

        cell = [matrix cellAtRow:i column:0];
        [cell setRepresentedObject:container];
        [cell setData:nil];
        [cell setDataProvider:self];
        i++;
    }
    [matrix sizeToCells];
    [matrix setNeedsDisplay];
}

- (void)provideDataForCell:(PieCell *)cell
{
    NSEnumerator *enumerator;
    id entity;
    id array;
    id container;

    if (!filter) return;

    if (!startTime) [self setStartTime:[filter startTime]];
    if (!endTime) [self setEndTime:[filter endTime]];

    container = [cell representedObject];
    NSAssert(container, @"Invalid container in StatViewer cell");

    array = [[StatArray alloc] initWithName:[NSString stringWithFormat:@"%@ %@", [[container entityType] description], [container description]]];

    switch ([[graphSelector selectedItem] tag]) {
        case 0:
            enumerator = [filter enumeratorOfEntityType:@"Activity State"
                                            inContainer:container
                                               fromTime:startTime
                                                 toTime:endTime];
            while(entity = [enumerator nextObject]) {
                StatValue *v;
                NSDate *entityStartTime = [filter startTimeForEntity:entity];
                NSDate *entityEndTime = [filter endTimeForEntity:entity];
                NSDate *firstTime = [startTime laterDate:entityStartTime];
                NSDate *lastTime = [endTime earlierDate:entityEndTime];
                double interval = [lastTime timeIntervalSinceDate:firstTime];
                v = [StatValue valueWithValue:interval
                                        color:[filter colorForEntity:entity]
                                         name:[filter nameForEntity:entity]];
                [array addObject:v];
            }
                break;

        case 1:
            enumerator = [filter enumeratorOfEntityType:@"Event"
                                            inContainer:container
                                               fromTime:startTime
                                                 toTime:endTime];
            while(entity = [enumerator nextObject]) {
                StatValue *v;
                v = [StatValue valueWithValue:1
                                        color:[filter colorForEntity:entity]
                                         name:[filter nameForEntity:entity]];
                [array addObject:v];
            }
                break;

        case 2:
            enumerator = [filter enumeratorOfEntityType:@"Communication"
                                            inContainer:[filter rootInstance]
                                               fromTime:startTime
                                                 toTime:endTime];
            while(entity = [enumerator nextObject]) {
                id size = [(A0Comm*)entity size];
                if (size) {
                    id source = [filter sourceContainerForEntity:entity];
                    id dest = [filter destContainerForEntity:entity];
                    StatValue *v;
                    if ([source isEqual: container]) {
                        v = [StatValue valueWithValue:[size doubleValue]
                                                color:[filter colorForEntity:entity]
                                                 name:[@"To " stringByAppendingString:[filter nameForEntity:dest]]];
                        [array addObject:v];
                    }
                    if ([dest isEqual: container]) {
                        v = [StatValue valueWithValue:[size doubleValue]
                                                color:[filter colorForEntity:entity]
                                                 name:[@"From " stringByAppendingString:[filter nameForEntity:source]]];
                        [array addObject:v];
                    }
                }
            }
                break;
    }

    [cell setData:array];
    [cell setGraphType:[[graphTypeSelector selectedItem] tag]];
    [cell setShowPercent:[[percentSelector selectedCell] tag]];
    [array release];
}

- (void)changeData:(id)sender
{
    [self invalidateValues];
}

- (void)changeGraph:(id)sender
{
    int rows, cols;
    PieCell *cell;
    int i;

    [matrix getNumberOfRows:&rows columns:&cols];
    for (i=0; i<rows; i++) {
        cell = [matrix cellAtRow:i column:0];
        [cell setGraphType:[[graphTypeSelector selectedItem] tag]];
        [cell setShowPercent:[[percentSelector selectedCell] tag]];
    }
    [matrix setNeedsDisplay];
}

- (void)changeInitialAngle:(id)sender;
{
    [[matrix cells] makeObjectsPerformSelector:@selector(setInitialAngle:)
                                    withObject:sender];
    [matrix display];
}

- (void)windowDidResize:(NSNotification *)notification
{
        NSSize cellSize;
        NSRect matrixFrame;

#if 1
        cellSize = [matrix cellSize];
        matrixFrame = [matrix frame];
        cellSize.width = NSWidth(matrixFrame);
        [matrix setCellSize:cellSize];
        [matrix sizeToCells];
        [matrix setNeedsDisplay];
#else
        [matrix setAutosizesCells:YES];
        [matrix setValidateSize:NO];
#endif
        [matrix setNeedsDisplay];
        cellSize = [matrix cellSize];

        [[NSUserDefaults standardUserDefaults] setObject:NSStringFromSize(cellSize) forKey:@"StatViewerCellSize"];
}

- (void)print:(id)sender
{
//    [[NSPrintOperation printOperationWithView:[window contentView]] runOperation]; 
    [window print:sender];
}
@end
