  /*
   * VirtualThread.m
   *
   * Component that re-numbers the threads in a trace, reusing
   * the numbers as the threads die.
   * Eve component for the Scope visualising tool.
   *
   * 19960514 BS  creation
   */

#import "VirtualThread.h"
#import "FoundationAdditions.h"


@interface MultiEnumerator : NSEnumerator
{
    NSMutableArray *origEnums;
}

+ (MultiEnumerator *)enumeratorWithEnumerators:(NSArray *)array;
- (id)initWithEnumerators:(NSArray *)array;
- (id)nextObject;
@end
@implementation MultiEnumerator
+ (MultiEnumerator *)enumeratorWithEnumerators:(NSArray *)array;
{
    return [[[self alloc] initWithEnumerators:array] autorelease];
}
- (id)initWithEnumerators:(NSArray *)array;
{
    [super init];
    origEnums = [array mutableCopy];
    return self;
}
- (void)dealloc
{
    [origEnums release];
    [super dealloc];
}
- (id)nextObject;
{
    id obj = nil;

    while (obj == nil) {

        if ([origEnums count] == 0)
            return nil;

        obj = [[origEnums objectAtIndex:0] nextObject];
        if (obj)
            return obj;

        [origEnums removeObjectAtIndex:0];

    }
    return nil; // shuts up compiler
}
@end



@implementation VirtualThread

// methods for reutilisation of thread slots on the graph,
// by changing thread numbers.

- (id)init
{
    [super init];

    nodes = [[NSMutableDictionary alloc] init];

    // load the bundle. it initializes view.
    if (![NSBundle loadNibNamed:@"VirtualThread" owner:self])
        NSRunAlertPanel(NSStringFromClass([self class]),
                        @"Couldn't load interface file",
                        nil, nil, nil);

    defaultFilteringMode = [filteringModeSelector selectedTag];

    // view is an NSBox. we need its contents.
    view = [view contentView];
    [[view window] autorelease];
    [view removeFromSuperview];

    [[NSNotificationCenter defaultCenter] postNotificationName:@"PajeRegisterFilterNotification" object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:
                        NSStringFromClass([self class]), @"Name",
                        view, @"View", nil]];

    return self;
}

- (void)dealloc
{
    [trace release];
    [nodes release];
    [lastAnalysedTime release];
    [super dealloc];
}

- (void)reset
{
    [trace release];
    trace = nil;
    [lastAnalysedTime release];
    lastAnalysedTime = nil;
    [nodes removeAllObjects];
    [nodeSelector removeAllItems];
    [nodeSelector addItemWithTitle:@"All"];
    [filteringModeSelector selectCellWithTag:defaultFilteringMode];
}

- (void)changeSelectedNode:(id)sender
{
    id nodeTitle = [sender titleOfSelectedItem];
    id node = [nodes objectForKey:nodeTitle];
    filterType filteringMode;

    if (!node)
        node = [nodes objectForKey:[NSNumber numberWithInt:[nodeTitle intValue]]];

    filteringMode = [node filteringMode];
    [filteringModeSelector selectCellWithTag:filteringMode];
}

- (void)changeFilteringMode:(id)sender
{
    id nodeTitle = [nodeSelector titleOfSelectedItem];
    id node;
    filterType filteringMode = [sender selectedTag];

    if ([nodeTitle isEqual:@"All"]) {
        id nodeEnum = [nodes objectEnumerator];
        defaultFilteringMode = filteringMode;
        while (node = [nodeEnum nextObject])
            [node setFilteringMode:filteringMode];
    } else {
        node = [nodes objectForKey:nodeTitle];
        if (!node)
            node = [nodes objectForKey:[NSNumber numberWithInt:[nodeTitle intValue]]];
        [node setFilteringMode:filteringMode];
    }

    // inform our successor that the trace has changed
    [self outputEntity:self];
}


- (void)analyseUntilTime:(NSDate *)time
{
    NSEnumerator *nodeEnum, *threadEnum;
    id nodeId;
    id thread;
    VTNode *node;
    BOOL starting;
    BOOL filterChanged = NO;
    id threadId, service;
    NSDate *t1;
    NSTimeInterval s1, s2;

    t1 = [NSDate date];
    if ((nil != trace)
        && (nil != lastAnalysedTime)
        && ([lastAnalysedTime earlierThanDate:time])) {

        NSMutableSet *unexistantNodes = [[NSMutableSet alloc] initWithArray:[nodes allKeys]];

        //nodeEnum = [trace nodeEnumerator];
        //*** we need a way of getting a hierarchy
        nodeEnum = [trace containerEnumeratorForHierarchy:[NSArray arrayWithObject:@"Node"]
                                         inSupercontainer:[NSArray array]];
        while (nodeId = [nodeEnum nextObject]) {
            [unexistantNodes removeObject:nodeId];

            node = [nodes objectForKey:nodeId];
            if (nil == node) {
                node = [VTNode node];
                [node setFilteringMode:defaultFilteringMode];
                [nodes setObject:node forKey:nodeId];
                [nodeSelector addItemWithTitle:[nodeId description]];
            }

            /***entityEnum = [trace enumeratorOfEntities:@"PajeEvent"
                                              inNode:nodeId
                                            fromTime:lastAnalysedTime
                                              toTime:time];*/
            threadEnum = [trace containerEnumeratorForHierarchy:[NSArray arrayWithObjects:@"Node", @"Thread", nil]
                                               inSupercontainer:nodeId];
            while (thread = [threadEnum nextObject]) {
                    threadId = thread;
                    service = [thread service];
                    if (1/*starting*/) {
                        [node startThread:threadId service:service];
                        filterChanged = YES;
                    } else {
                        [node endThread:threadId service:service];
                    }
            }
        }

        // remove nodes that don't exist in trace from nodes and nodeSelector
        nodeEnum = [unexistantNodes objectEnumerator];
        while (nodeId = [nodeEnum nextObject]) {
            [nodes removeObjectForKey:nodeId];
            [nodeSelector removeItemWithTitle:[nodeId description]];
        }
        [unexistantNodes release];

        // update lastAnalysedTime
        [lastAnalysedTime release];
        lastAnalysedTime = [time retain];
//        NSLog(@"Analysed until %f", [time timeIntervalSinceReferenceDate]);

        // send the modified trace out
//        s1 = -[t1 timeIntervalSinceNow];
        if (filterChanged)
            [self outputEntity:self];
//        s2 = -[t1 timeIntervalSinceNow];
//        NSLog(@"VT analysed %f+%f=%fs", s1, s2-s1, s2);
    }
}

- (void)inputEntity:(id)entity
{
    Trace *newTrace = entity;

//    NSLog(@"in -[%@ %@]", NSStringFromClass([self class]), NSStringFromSelector(_cmd));
    if (trace != newTrace) {
        // let's forget everything
        [self reset];
        trace = [newTrace retain];
//***        lastAnalysedTime = [[trace startTime] retain];
    }

    [lastAnalysedTime release];
    lastAnalysedTime = nil;
    //    if (!lastAnalysedTime)//*** BUG KLUDGE Trace should't give stupid times when empty
        if ([[trace endTime] timeIntervalSinceDate:[trace startTime]] > 0.)//***
            lastAnalysedTime = [[trace startTime] retain];//***
//    if (lastAnalysedTime)//***
//        [self analyseUntilTime:[trace endTime]];

            [hierarchyToSupercontainers release];
            [containerToNewContainer release];
            hierarchyToSupercontainers = containerToNewContainer = nil;

            [self outputEntity:self];
}

- (NSEnumerator *)enumeratorOfEntityType:(id)entityType
                             inContainer:(id)container
                                fromTime:(NSDate *)start
                                  toTime:(NSDate *)end
{
    if ([container isKindOfClass:[NSArray class]]) {
        NSMutableArray *enumArray = [NSMutableArray array];
        NSEnumerator *compEnum = [container objectEnumerator];
        id compContainer;

        while (compContainer = [compEnum nextObject]) {
            [enumArray addObject:[trace enumeratorOfEntityType:entityType
                                                   inContainer:compContainer
                                                      fromTime:start
                                                        toTime:end]];
        }
        return [MultiEnumerator enumeratorWithEnumerators:enumArray];
    }
    return [trace enumeratorOfEntityType:entityType
                             inContainer:container
                                fromTime:start
                                  toTime:end];
}


- (NSEnumerator *)containerEnumeratorForHierarchy:(id)hierarchy
                                 inSupercontainer:(id)supercontainer
{
    NSMutableDictionary *superToContainers;
    NSMutableArray *containers;

    if (nil == hierarchyToSupercontainers)
        hierarchyToSupercontainers = [[NSMutableDictionary alloc] init];
    
    superToContainers = [hierarchyToSupercontainers objectForKey:hierarchy];
    if (nil == superToContainers) {
        superToContainers = [NSMutableDictionary dictionary];
        [hierarchyToSupercontainers setObject:superToContainers forKey:hierarchy];
    }

    containers = [superToContainers objectForKey:supercontainer];
    if (nil == containers) {
        NSEnumerator *enumerator;
        id container;

        containers = [NSMutableArray array];
        [superToContainers setObject:containers forKey:supercontainer];

        if (([supercontainer isKindOfClass:[NSArray class]])
            && ([supercontainer count] > 0)) {
            NSMutableArray *enumArray = [NSMutableArray array];
            NSEnumerator *compEnum = [supercontainer objectEnumerator];
            id compContainer;
            while (compContainer = [compEnum nextObject]) {
                [enumArray addObject:
                    [trace containerEnumeratorForHierarchy:hierarchy
                                          inSupercontainer:compContainer]];
            }
            enumerator = [MultiEnumerator enumeratorWithEnumerators:enumArray];
        } else
            enumerator = [trace containerEnumeratorForHierarchy:hierarchy
                                               inSupercontainer:supercontainer];
        while (container = [enumerator nextObject]) {
            NSEnumerator *newEnum = [containers objectEnumerator];
            NSMutableArray *newContainer;

            while (newContainer = [newEnum nextObject]) {
                NSEnumerator *compEnum = [newContainer objectEnumerator];
                id compContainer;
                BOOL conflicts = NO;
                while (compContainer = [compEnum nextObject]) {
                    if (!([[container endTime] earlierThanDate:[compContainer startTime]]
                          || [[compContainer endTime] earlierThanDate:[container startTime]])) {
                        conflicts = YES;
                        break;
                    }
                }
                if (conflicts == NO)
                    break;
            }
            if (nil == newContainer) {
                newContainer = [NSMutableArray array];
                [containers addObject:newContainer];
            }
            [newContainer addObject:container];
            if (!containerToNewContainer)
                containerToNewContainer = [[NSMutableDictionary alloc] init];
            [containerToNewContainer setObject:newContainer forKey:container];
        }
    }

    return [containers objectEnumerator];
}

- (id)newContainerForContainer:(id)container;
{
    return [containerToNewContainer objectForKey:container];
}

- (id)containerForEntity:(id<PajeEntity>)entity;
{
    return [self newContainerForContainer:[trace containerForEntity:entity]];
}

- (id)sourceContainerForEntity:(id<PajeLink>)entity;
{
    return [self newContainerForContainer:[trace sourceContainerForEntity:entity]];
}

- (id)destContainerForEntity:(id<PajeLink>)entity;
{
    return [self newContainerForContainer:[trace destContainerForEntity:entity]];
}



- (int)slotForThread:(id)threadId
             service:(id)service
                node:(id)nodeId
{
    VTNode *node = [nodes objectForKey:nodeId];

    if (nil != node) {
        return [node slotForThread:threadId service:service];
    } else {
        NSLog(@"%@: slot for thread %@ in unknown node %@",
              NSStringFromClass([self class]), threadId, nodeId);
        return [threadId intValue];
    }
}



- (int)threadCountForNode:(id)nodeId
{
    VTNode *node = [nodes objectForKey:nodeId];

    if (nil != node) {
        if ([node filteringMode] == NO_FILTERING)
            return [trace threadCountForNode:nodeId];
        else
            return [node threadCount];
    } else {
        NSLog(@"%@: threadCount for unknown node %@",
              NSStringFromClass([self class]), nodeId);
        return [trace threadCountForNode:nodeId];
    }
}


- (NSEnumerator *)enumeratorOfEntities:(id)entityType
                                inNode:(id)node
                              fromTime:(NSDate *)start
                                toTime:(NSDate *)end
{
    if (trace) {
        [self analyseUntilTime:end];

        return [trace enumeratorOfEntities:entityType
                                    inNode:node
                                  fromTime:start
                                    toTime:end];
    } else {
        return nil;
    }
}

- (NSEnumerator *)enumeratorOfEntities:(id)entityType
                              fromTime:(NSDate *)start
                                toTime:(NSDate *)end
{
    if (trace) {
        [self analyseUntilTime:end];

        return [trace enumeratorOfEntities:entityType
                                  fromTime:start
                                    toTime:end];
    } else {
        return nil;
    }
}

- (void)getNode:(id *)node
         thread:(id *)thread
      startTime:(NSDate **)start
        endTime:(NSDate **)end
          color:(NSColor **)color
       ofEntity:(id<PajeState>)entity
{
    [trace getNode:node
            thread:thread
         startTime:start
           endTime:end
             color:color
          ofEntity:entity];
    
    //KLUDGE for semaphores and mutexes
    if ([[entity entityType] isEqual:@"PajeSemaState"]
        || [[entity entityType] isEqual:@"Node Activity State"])
        return;
    //KLUDGE there should be a distinction between thread id and thread slot
    if (0/****thread*/) {
        *thread = [NSNumber numberWithInt:[self slotForThread:*thread
                                                      service:[entity service]
                                                         node:*node]];
    }
    if (thread /*****/&& ([[nodes objectForKey:*node] filteringMode] != NO_FILTERING)) {
        *thread = [NSNumber numberWithInt:[self slotForThread:[entity thread]//*** *thread
                                                      service:[entity service]
                                                         node:*node]];
    }
}

- (void)getSourceNode:(id *)sourceNode
             destNode:(id *)destNode
         sourceThread:(id *)sourceThread
           destThread:(id *)destThread
            startTime:(NSDate **)start
              endTime:(NSDate **)end
                color:(NSColor **)color
             ofEntity:(id<PajeComm>)entity
{
    [trace getSourceNode:sourceNode
                destNode:destNode
            sourceThread:sourceThread
              destThread:destThread
               startTime:start
                 endTime:end
                   color:color
                ofEntity:entity];
    //KLUDGE for semaphores and mutexes
    if ([[entity entityType] isEqual:@"PajeSemaState"]
        || [[entity entityType] isEqual:@"Node Activity State"])
        return;
    //KLUDGE there should be a distinction between thread id and thread slot
    if (0/****sourceThread*/) {
        *sourceThread = [NSNumber numberWithInt:
            [self slotForThread:*sourceThread
                        service:[entity sourceService]
                           node:*sourceNode]];
    }
    if (0/****destThread*/) {
        *destThread = [NSNumber numberWithInt:
            [self slotForThread:*destThread
                        service:[entity destService]
                           node:*destNode]];
    }
    if (sourceThread /*****/&& ([[nodes objectForKey:*sourceNode] filteringMode] != NO_FILTERING)) {
        *sourceThread = [NSNumber numberWithInt:
            [self slotForThread:[entity sourceThread]
                        service:[entity sourceService]
                           node:*sourceNode]];
    }
    if (destThread /*****/&& ([[nodes objectForKey:*destNode] filteringMode] != NO_FILTERING)) {
        *destThread = [NSNumber numberWithInt:
            [self slotForThread:[entity destThread]
                        service:[entity destService]
                           node:*destNode]];
    }
}


// to be a PajeComponent.. *** shouldn't this be in a superclass?
- (void)setOutputComponent:(id)component selector:(SEL)sel
{
    outputComponent = component;
    outputSelector = sel;
}

- (void)outputEntity:(id)entity
{
    [outputComponent performSelector:outputSelector withObject:entity];
}


//
// Forwards other messages to trace object
//

- (BOOL)respondsToSelector:(SEL)aSelector
{
    return ([super respondsToSelector:aSelector]
            || [trace respondsToSelector:aSelector]);
}

- (void)forwardInvocation:(NSInvocation *)invocation
{
    [invocation invokeWithTarget:trace];
}

- (NSMethodSignature *)methodSignatureForSelector:(SEL)sel {
    NSMethodSignature *result = [super methodSignatureForSelector:sel];
    if (nil == result)
        result = [trace methodSignatureForSelector:sel];
    return result;
}

@end
